"""

Asynchroneous extension of Julabo

"""

import fluidlab.instruments.chiller.julabo as fl_ju
from pymanip.aioinstruments.aiodrivers import AsyncDriver
from pymanip.aioinstruments.aiofeatures import (
    AsyncFloatValue,
    AsyncBoolValue,
    AsyncValue,
    AsyncIntValue,
)


class AsyncJulabo(AsyncDriver, fl_ju.Julabo):
    async def __aenter__(self):
        await super().__aenter__()
        identification = await self.interface.aquery("version")
        print("identification =", repr(identification))
        return self


afeatures = [
    AsyncFloatValue(
        "setpoint",
        channel_argument=True,
        check_instrument_value=False,
        pause_instrument=0.5,
        command_get="in_sp_{channel:02d}",
        command_set="out_sp_{channel:02d} {value:.1f}",
    ),
    AsyncBoolValue(
        "onoff",
        command_set="out_mode_05",
        pause_instrument=0.5,
        check_instrument_value=False,
    ),
    AsyncFloatValue("temperature", pause_instrument=0.75, command_get="in_pv_00"),
    AsyncIntValue(
        "setpoint_channel",
        check_instrument_value=False,
        pause_instrument=0.5,
        command_get="in_mode_01",
        command_set="out_mode_01",
    ),
    AsyncValue("status", pause_instrument=0.5, command_get="status"),
]

AsyncJulabo._build_class_with_features(afeatures)



"""Andor Camera module (:mod:`pymanip.video.andor.camera`)
==========================================================

This module implement the :class:`Andor_Camera` class, as a subclass
of the :class:`pymanip.video.Camera` base class. It uses the
third-party :mod:`pyAndorNeo` module.

.. autoclass:: Andor_Camera
   :members:
   :private-members:

"""

import time
import asyncio
import ctypes
import struct

import numpy as np
from pymanip.video import MetadataArray, Camera, CameraTimeout

import AndorNeo.SDK3Cam as SDK3Cam
import AndorNeo.SDK3 as SDK3

MODE_CONTINUOUS = 1
MODE_SINGLE_SHOT = 0

validROIS = [
    (2592, 2160, 1, 1),
    (2544, 2160, 1, 25),
    (2064, 2048, 57, 265),
    (1776, 1760, 201, 409),
    (1920, 1080, 537, 337),
    (1392, 1040, 561, 601),
    (528, 512, 825, 1033),
    (240, 256, 953, 1177),
    (144, 128, 1017, 1225),
]


def parse_metadata(buf, verbose=False):
    """
    Parse metadata at the end of buffer, happened when EnableMetadata boolean
    feature is true.(page 77 doc SDK3 manual)

    Metadata format: Frame data; CID; Length
    FrameData (CID=0), FPGA Ticks (CID=1), FrameInfo (CID=7)
    """
    LENGTH_FIELD_SIZE = 4
    CID_FIELD_SIZE = 4
    # TIMESTAMP_FIELD_SIZE = 8

    n = buf.size
    for ind in range(3):
        length = buf[n - LENGTH_FIELD_SIZE : n]
        cid = buf[n - (CID_FIELD_SIZE + LENGTH_FIELD_SIZE) : n - LENGTH_FIELD_SIZE]
        # if verbose:
        #    print('length =', length)
        #    print('cid =', cid)
        # length = length[0]+(2**8)*length[1]+(2**16)*length[2]+\
        # (2**32)*length[3]-CID_FIELD_SIZE
        # cid = cid[0]+(2**8)*cid[1]+(2**16)*cid[2]+(2**32)*cid[3]
        length = struct.unpack("<L", length)[0] - CID_FIELD_SIZE
        cid = struct.unpack("<L", cid)[0]
        data = buf[
            n
            - (CID_FIELD_SIZE + LENGTH_FIELD_SIZE + length) : n
            - (CID_FIELD_SIZE + LENGTH_FIELD_SIZE)
        ]
        if verbose:
            print("length =", length)
            print("cid =", cid)
            print("data =", data)
        if cid == 1:
            return struct.unpack("<Q", data)[0]
            # return sum([(256**i)*b for i, b in enumerate(data)])

        n -= CID_FIELD_SIZE + LENGTH_FIELD_SIZE + length


class Andor_Camera(Camera):
    """Concrete :class:`pymanip.video.Camera` class for Andor camera.

    :param camNum: camera number, defaults to 0.
    :type camNum: int, optional
    """

    def __init__(self, camNum=0):
        """Constructor method
        """

        self.camNum = camNum

        # Auto properties (bound in SDK3Cam __init__)
        self.CameraAcquiring = SDK3Cam.ATBool()
        self.SensorCooling = SDK3Cam.ATBool()
        self.MetadataEnable = SDK3Cam.ATBool()
        self.TimestampClock = SDK3Cam.ATInt()
        self.TimestampClockFrequency = SDK3Cam.ATInt()

        self.AcquisitionStart = SDK3Cam.ATCommand()
        self.AcquisitionStop = SDK3Cam.ATCommand()

        self.CycleMode = SDK3Cam.ATEnum()
        self.ElectronicShutteringMode = SDK3Cam.ATEnum()
        self.FanSpeed = SDK3Cam.ATEnum()
        self.PreAmpGainChannel = SDK3Cam.ATEnum()
        self.PixelEncoding = SDK3Cam.ATEnum()
        self.PixelReadoutRate = SDK3Cam.ATEnum()
        self.PreAmpGain = SDK3Cam.ATEnum()
        self.PreAmpGainSelector = SDK3Cam.ATEnum()
        self.TriggerMode = SDK3Cam.ATEnum()

        self.AOIHeight = SDK3Cam.ATInt()
        self.AOILeft = SDK3Cam.ATInt()
        self.AOITop = SDK3Cam.ATInt()
        self.AOIWidth = SDK3Cam.ATInt()
        self.AOIStride = SDK3Cam.ATInt()
        self.FrameCount = SDK3Cam.ATInt()
        self.ImageSizeBytes = SDK3Cam.ATInt()
        self.SensorHeight = SDK3Cam.ATInt()
        self.SensorWidth = SDK3Cam.ATInt()

        self.CameraModel = SDK3Cam.ATString()
        self.SerialNumber = SDK3Cam.ATString()

        self.ExposureTime = SDK3Cam.ATFloat()
        self.FrameRate = SDK3Cam.ATFloat()
        self.SensorTemperature = SDK3Cam.ATFloat()
        self.TargetSensorTemperature = SDK3Cam.ATFloat()

        self.Overlap = SDK3Cam.ATBool()
        self.SpuriousNoiseFilter = SDK3Cam.ATBool()

        self.CameraDump = SDK3Cam.ATCommand()
        self.SoftwareTrigger = SDK3Cam.ATCommand()

        self.TemperatureControl = SDK3Cam.ATEnum()
        self.TemperatureStatus = SDK3Cam.ATEnum()
        self.SimplePreAmpGainControl = SDK3Cam.ATEnum()
        self.BitDepth = SDK3Cam.ATEnum()

        self.ActualExposureTime = SDK3Cam.ATFloat()
        self.BurstRate = SDK3Cam.ATFloat()
        self.ReadoutTime = SDK3Cam.ATFloat()

        self.AccumulateCount = SDK3Cam.ATInt()
        self.BaselineLevel = SDK3Cam.ATInt()
        self.BurstCount = SDK3Cam.ATInt()
        self.LUTIndex = SDK3Cam.ATInt()
        self.LUTValue = SDK3Cam.ATInt()

        self.ControllerID = SDK3Cam.ATString()
        self.FirmwareVersion = SDK3Cam.ATString()

        # Initialisation
        self.handle = SDK3.Open(self.camNum)
        for name, var in self.__dict__.items():
            if isinstance(var, SDK3Cam.ATProperty):
                var.connect(self.handle, name)

        # set some initial parameters
        # self.FrameCount.setValue(1) #only for fixed mode?
        self.CycleMode.setString("Continuous")
        self.SimplePreAmpGainControl.setString("11-bit (low noise)")
        self.PixelEncoding.setString("Mono12Packed")  # Mono12Packed Mono16

        # Camera object properties
        self.FrameRate.setValue(self.FrameRate.max())
        self.name = (
            "Andor " + self.CameraModel.getValue() + " " + self.SerialNumber.getValue()
        )
        print(self.name)

    def close(self):
        """This method closes the Camera handle.
        """
        SDK3.Close(self.handle)

    def __exit__(self, type_, value, cb):
        """Context manager exit method.
        """
        super(Andor_Camera, self).__exit__(type_, value, cb)
        self.close()

    def set_exposure_time(self, seconds):
        """This method sets the camera exposure time

        :param seconds: exposure in seconds
        :type seconds: float

        """
        self.ExposureTime.setValue(seconds)

    def get_exposure_time(self):
        """This method returns the exposure time in seconds
        """
        return self.ExposureTime.getValue()

    def acquisition_oneshot(self, timeout=1.0):
        """Concrete implementation of :meth:`pymanip.video.Camera.acquisition_oneshot` for the Andor camera.
        """

        # Make sure no acquisition is running & flush
        if self.CameraAcquiring.getValue():
            self.AcquisitionStop()
        SDK3.Flush(self.handle)
        self.MetadataEnable.setValue(True)
        pc_clock = time.time()
        timestamp_clock = self.TimestampClock.getValue()
        timestamp_frequency = self.TimestampClockFrequency.getValue()

        # Init buffer & queue
        bufSize = self.ImageSizeBytes.getValue()
        buf = np.empty(bufSize, "uint8")
        SDK3.QueueBuffer(
            self.handle, buf.ctypes.data_as(SDK3.POINTER(SDK3.AT_U8)), buf.nbytes
        )

        # Start acquisition
        self.AcquisitionStart()
        print("Start acquisition at framerate:", self.FrameRate.getValue())

        try:
            # Wait for buffer
            exposure_ms = self.ExposureTime.getValue() * 1000
            framerate_ms = 1000 / self.FrameRate.getValue()
            timeout_ms = int(max((2 * exposure_ms, 2 * framerate_ms, 1000)))

            pData, lData = SDK3.WaitBuffer(self.handle, timeout_ms)

            # Convert buffer into numpy image
            rbuf, cbuf = self.AOIWidth.getValue(), self.AOIHeight.getValue()
            img = np.empty((rbuf, cbuf), np.uint16)
            xs, ys = img.shape[:2]
            a_s = self.AOIStride.getValue()
            dt = self.PixelEncoding.getString()
            ticks = parse_metadata(buf)
            ts = (ticks - timestamp_clock) / timestamp_frequency + pc_clock
            SDK3.ConvertBuffer(
                buf.ctypes.data_as(ctypes.POINTER(ctypes.c_uint8)),
                img.ctypes.data_as(ctypes.POINTER(ctypes.c_uint8)),
                xs,
                ys,
                a_s,
                dt,
                "Mono16",
            )
        finally:
            self.AcquisitionStop()

        return MetadataArray(
            img.reshape((cbuf, rbuf), order="C"), metadata={"timestamp": ts}
        )

    async def acquisition_async(
        self,
        num=np.inf,
        timeout=None,
        raw=False,
        initialising_cams=None,
        raise_on_timeout=True,
    ):
        """Concrete implementation of :meth:`pymanip.video.Camera.acquisition_async` for the Andor camera.

        .. todo::
            add support for initialising_cams
        """

        loop = asyncio.get_event_loop()

        # Make sure no acquisition is running & flush
        if self.CameraAcquiring.getValue():
            self.AcquisitionStop()
        SDK3.Flush(self.handle)
        self.buffer_queued = False

        # Set acquisition mode
        self.CycleMode.setString("Continuous")
        # self.FrameRate.setValue(float(framerate))
        self.MetadataEnable.setValue(True)
        pc_clock = time.time()
        timestamp_clock = self.TimestampClock.getValue()
        timestamp_frequency = self.TimestampClockFrequency.getValue()
        print("ts clock =", timestamp_clock)
        print("ts freq =", timestamp_frequency)

        # Init buffers
        bufSize = self.ImageSizeBytes.getValue()
        buf = np.empty(bufSize, "uint8")
        rbuf, cbuf = self.AOIWidth.getValue(), self.AOIHeight.getValue()
        img = np.empty((rbuf, cbuf), np.uint16)
        xs, ys = img.shape[:2]
        a_s = self.AOIStride.getValue()
        dt = self.PixelEncoding.getString()
        print("Original pixel encoding:", dt)

        # Start acquisition
        self.AcquisitionStart()
        print("Started acquisition at framerate:", self.FrameRate.getValue())
        print("Exposure time is {:.1f} ms".format(self.ExposureTime.getValue() * 1000))
        if timeout is None:
            exposure_ms = self.ExposureTime.getValue() * 1000
            framerate_ms = 1000 / self.FrameRate.getValue()
            timeout_ms = int(max((2 * exposure_ms, 2 * framerate_ms, 1000)))
            timeout = timeout_ms

        try:
            count = 0
            while count < num:
                if not self.buffer_queued:
                    SDK3.QueueBuffer(
                        self.handle,
                        buf.ctypes.data_as(SDK3.POINTER(SDK3.AT_U8)),
                        buf.nbytes,
                    )
                    self.buffer_queued = True
                try:
                    pData, lData = await loop.run_in_executor(
                        None, SDK3.WaitBuffer, self.handle, timeout
                    )
                except Exception:
                    if raise_on_timeout:
                        raise CameraTimeout()
                    else:
                        stop_signal = yield None
                        if stop_signal:
                            break
                        else:
                            continue
                # Convert buffer and yield image
                SDK3.ConvertBuffer(
                    buf.ctypes.data_as(ctypes.POINTER(ctypes.c_uint8)),
                    img.ctypes.data_as(ctypes.POINTER(ctypes.c_uint8)),
                    xs,
                    ys,
                    a_s,
                    dt,
                    "Mono16",
                )
                ticks = parse_metadata(buf)
                ts = (ticks - timestamp_clock) / timestamp_frequency + pc_clock
                # if count == 5:
                #    print('image min max:', np.min(img), np.max(img))
                # if count < 10:
                #    print('FPGA ticks =', ticks)
                #    print('Timestamp =', ts)
                self.buffer_queued = False
                stop_signal = yield MetadataArray(
                    img.reshape((cbuf, rbuf), order="C"),
                    metadata={"counter": count, "timestamp": ts},
                )
                count = count + 1
                if stop_signal:
                    break

        finally:
            self.AcquisitionStop()
        if stop_signal:
            yield True


if __name__ == "__main__":
    import matplotlib.pyplot as plt

    with Andor_Camera() as ac:
        ac.set_exposure_time(1e-3)
        img = ac.acquisition_oneshot()
    plt.figure()
    plt.imshow(img, cmap="gray")
    plt.show()



"""Asynchronous extension of :class:`~fluidlab.interfaces.gpib_inter.GPIBInterface` (:mod:`pymanip.interfaces.aiogpib`)
=======================================================================================================================

This module defines :class:`AsyncGPIBInterface` as a subclass of :class:`fluidlab.interfaces.gpib_inter.GPIBInterface` and
:class:`pymanip.interface.aiointer.AsyncQueryInterface`.

.. autoclass:: AsyncGPIBInterface
   :members:
   :private-members:

"""

import asyncio
import fluidlab.interfaces.gpib_inter as fl_gpib
from pymanip.interfaces.aiointer import AsyncQueryInterface

GPIBLock = asyncio.Lock()


class AsyncGPIBInterface(fl_gpib.GPIBInterface, AsyncQueryInterface):
    """This class is an asynchronous extension of :class:`fluidlab.interfaces.gpib_inter.GPIBInterface`.
    It inherits all its methods from the parent classes.
    """

    def __init__(self, board_adress, instrument_adress):
        """Constructor method
        """
        super().__init__(board_adress, instrument_adress)
        # replace per-device lock by global GPIB lock
        self.lock = GPIBLock

    async def await_for_srq(self, timeout=None):
        """This co-routine method acquires the interface lock and run wait_for_srq in an
        executor
        """
        async with self.lock:
            await self.loop.run_in_executor(None, self.wait_for_srq, timeout)



"""Asynchronous Acquisition Card (:mod:`pymanip.aiodaq`)
========================================================

This module defines an abstract base class for asynchronous communication
with acquisition cards.
This is used by the live oscilloscope command line tool.

Concrete implementations are:

- :class:`pymanip.aiodaq.daqmx.DAQmxSystem` for NI DAQmx cards;

- :class:`pymanip.aiodaq.scope.ScopeSystem` for NI Scope cards.

In principle, other library bindings could be implemented.

.. autoclass:: TerminalConfig

.. autoclass:: TriggerConfig

.. autoclass:: TimeoutException

.. autoclass:: AcquisitionCard
   :members:
   :private-members:

"""

import asyncio
from enum import IntEnum

from pymanip.asynctools import synchronize_function


class TerminalConfig(IntEnum):
    RSE = 0
    NRSE = 1
    Diff = 2
    PseudoDiff = 3


class TriggerConfig(IntEnum):
    EdgeRising = 0


class TimeoutException(Exception):
    pass


class AcquisitionCard:
    """Base class for all acquisition cards.
    The constructor takes no argument. Channels
    are added using the :meth:`~pymanip.aiodaq.daqmx.DAQmxSystem.add_channel` method,
    and the clock is configured with the
    :meth:`~pymanip.aiodaq.daqmx.DAQmxSystem.configure_clock` method.
    """

    def __init__(self):
        """Constructor method
        """
        self.channels = []
        self.actual_ranges = []
        self.running = False
        self.last_read = 0
        self.sample_rate = None
        self.samples_per_chan = 1
        self.loop = asyncio.get_event_loop()

    def __enter__(self):
        """Context manager enter method
        """
        return self

    def __exit__(self, type_, value, cb):
        """Context manager exit method
        """
        self.close()

    @property
    def samp_clk_max_rate(self):
        """Maximum sample clock rate
        """
        return 0

    def possible_trigger_channels(self):
        """This method returns the list of channels that can be used as trigger.
        """
        return self.channels

    def close(self):
        """This method closes the connection to the acquisition card.
        """
        raise NotImplementedError()

    def add_channel(self, channel_name, terminal_config, voltage_range):
        """This method adds a channel for acquisition.

        :param channel_name: the channel to add, e.g. "Dev1/ai0"
        :type channel_name: str
        :param terminal_config: the configuration of the terminal, i.e. RSE, NRSE, DIFFERENTIAL or PSEUDODIFFERENTIAL
        :type terminal_config: :class:`~pymanip.aiodaq.TerminalConfig`
        :param voltage_range: the voltage range for the channel (actual value may differ)
        :type voltage_range: float
        """
        raise NotImplementedError()

    def configure_clock(self, sample_rate, samples_per_chan):
        """This method configures the board clock for the acquisition.

        :param sample_rate: the clock frequency in Hz
        :type sample_rate: float
        :param samples_per_chan: number of samples to be read on each channel
        :type samples_per_chan: int
        """
        raise NotImplementedError()

    def configure_trigger(
        self,
        trigger_source=None,
        trigger_level=0,
        trigger_config=TriggerConfig.EdgeRising,
    ):
        """This method configures the trigger for the acquisition, i.e. internal trigger
        or triggered on one of the possible channels. The list of possible channels
        can be obtained from the
        :meth:`~pymanip.aiodaq.daqmx.DAQmxSystem.possible_trigger_channels` method.

        :param trigger_source: the channel to use for triggering, or None to disable external trigger (switch to Immeditate trigger). Defaults to None.
        :type trigger_source: str
        :param trigger_level: the voltage threshold for triggering
        :type trigger_level: float
        :param trigger_config: the kind of triggering, e.g. EdgeRising. Defaults to EdgeRising.
        :type trigger_config: :class:`pymanip.aiodaq.TriggerConfig`, optional
        """
        raise NotImplementedError()

    def start(self):
        """This method starts the acquisition
        """
        raise NotImplementedError()

    async def stop(self):
        """This asynchronous method aborts the acquisition
        """
        raise NotImplementedError()

    async def read(self, tmo=None):
        """This asynchronous method reads data from the acquisition card.
        """
        raise NotImplementedError()

    async def start_read_stop(self, tmo=None):
        """This asynchronous method starts the acquisition, reads the data, and
        stops the acquisition.

        :param tmo: timeout for reading, defaults to None
        :type tmo: float
        """
        self.start()
        data = await self.read(tmo)
        await self.stop()
        return data

    def read_sync(self, tmo=None):
        """This method is a synchronous wrapper around
        :meth:`~pymanip.aiodaq.AcquisitionCard.start_read_stop` method.
        It is a convenience facility for simple usage.
        """
        loop = asyncio.get_event_loop()
        data = loop.run_until_complete(self.start_read_stop(tmo))
        return data

    async def read_analog(
        self,
        resource_names,
        terminal_config,
        volt_min=None,
        volt_max=None,
        samples_per_chan=1,
        sample_rate=1,
        coupling_types="DC",
        output_filename=None,
        verbose=True,
    ):
        """This asynchronous method is a high-level method for simple case. It
        configures all the given channels, as well as the clock, then starts the
        acquisition, read the data, and stops the acquisition.

        It is essentially similar to :func:`pymanip.daq.DAQmx.read_analog`, except
        asynchronous and functionnal for other cards than DAQmx cards.

        :param resource_names: list of resources to read, e.g. ["Dev1/ai1", "Dev1/ai2"] for DAQmx cards, or name of the resource if only one channel is to be read.
        :type resource_names: list or str
        :param terminal_config: list of terminal configs for the channels
        :type terminal_config: list
        :param volt_min: minimum voltage expected on the channel
        :type volt_min: float
        :param volt_max: maximum voltage expected on the channel
        :type volt_max: float
        :param samples_per_chan: number of samples to read on each channel
        :type samples_per_chan: int
        :param sample_rate: frequency of the clock
        :type sample_rate: float
        :param coupling_type: coupling for the channel (e.g. AC or DC)
        :type coupling_type: str
        :param output_filename: filename for direct writting to the disk
        :type output_filename: str, optional
        :param verbose: verbosity level
        :type verbose: bool, optional
        """

        if isinstance(resource_names, str):
            resource_names = [resource_names]
        if isinstance(terminal_config, str):
            terminal_config = [terminal_config]
        try:
            terminal_config[0]
        except TypeError:
            terminal_config = [terminal_config]
        try:
            volt_min[0]
        except TypeError:
            volt_min = [volt_min]
        try:
            volt_max[0]
        except TypeError:
            volt_max = [volt_max]

        for chan_name, chan_tc, chan_vmin, chan_vmax in zip(
            resource_names, terminal_config, volt_min, volt_max
        ):
            volt_range = max([abs(chan_vmin), abs(chan_vmax)])
            self.add_channel(chan_name, chan_tc, volt_range)
        self.configure_clock(sample_rate, int(samples_per_chan))
        self.configure_trigger(None)
        data = await self.start_read_stop()
        return data

    def read_analog_sync(self, *args, **kwargs):
        """Synchronous wrapper around :meth:`pymanip.aiodaq.AcquisitionCard.read_analog`.
        """
        loop = asyncio.get_event_loop()
        data = loop.run_until_complete(self.read_analog(*args, **kwargs))
        return data



"""Read older Octave-based experiment session (:mod:`fluidlab.exp.octavesession`)
=================================================================================

Provides:

read_OctMI_session(sessionName, verbose=True)

"""

from struct import pack, unpack, error
from time import strftime, localtime
from functools import reduce

import numpy as np

from fluiddyn.util.terminal_colors import cprint


class OctaveReaderError(NameError):
    """ OctaveReader error """


def read_header(f, verbose):
    """ Reads Octave binary file header (one per file)

    Format is described in libinterp/corefcn/ls-oct-binary.cc
    recalled below.

    Header (one per file):
    =====================

    object               type            bytes
    ------               ----            -----
    magic number         string             10

    float format         integer             1

    """

    magic_number = f.read(10).decode("ascii")
    if verbose:
        print("magic number =", magic_number)
    if magic_number != "Octave-1-L":
        raise OctaveReaderError("UnknownFileFormat")
    float_format = ord(f.read(1))
    if verbose:
        print("float format =", float_format)
    return float_format


def read_scalar_var(f, verbose):
    """ Reads one variable of type 'scalar' from Octave binary file

    See: octave_scalar::load_binary from libinterp/octave-value/ov-scalar.cc
    from GNU Octave source files

    """

    tmp = ord(f.read(1))
    if tmp == 0:
        raise OctaveReaderError("ScalarFormatError")
    dtmp = unpack("d", f.read(8))[0]
    if verbose:
        print("      value:", dtmp)

    return dtmp


def read_matrix_var(f, verbose):
    """ Reads one variable of type 'matrix' from Octave binary file

    See: octave_matrix::load_binary from libinterp/octave-value/ov-re-mat.cc
    from GNU Octave source files

    """

    mdims = unpack("i", f.read(4))[0]
    if verbose:
        print("      mdims:", mdims)
    if mdims == 0:
        raise OctaveReaderError("MatrixFormatError")
    if mdims < 0:
        mdims = -mdims
        dv = unpack("i" * mdims, f.read(4 * mdims))
        if verbose:
            print("         dv:", dv)
        tmp = ord(f.read(1))
        if tmp == 0:
            raise OctaveReaderError("MatrixFormatError")
        num_elems = reduce(lambda x, y: x * y, dv)
        if verbose:
            print("  num_elems:", num_elems)
        fortran_vec = unpack("d" * num_elems, f.read(8 * num_elems))
        var = np.array(fortran_vec).reshape(dv, order="F")
        if verbose:
            print("     matrix:", var.shape)
        if mdims == 2 and dv[0] == 1:
            var = var.flatten()
    else:
        nr = mdims
        nc = unpack("i", f.read(4))[0]
        tmp = ord(f.read(1))
        if tmp == 0:
            raise OctaveReaderError("MatrixFormatError")
        dv = (nr, nc)
        num_elems = nr * nc
        if verbose:
            print("  num_elems:", num_elems)
        fortran_vec = unpack("d" * num_elems, f.read(8 * num_elems))
        var = np.array(fortran_vec).reshape(dv, order="F")
        if verbose:
            print("     matrix:", var.shape)

    return var


def read_complex_matrix_var(f, verbose):
    """ Reads one variable of type 'complex matrix' from Octave binary file

    See: octave_complex_matrix::load_binary from libinterp/octave-value/ov-cx-mat.cc
    from GNU Octave source files

    """

    mdims = unpack("i", f.read(4))[0]
    if mdims < 0:
        mdims = -mdims
        dv = unpack("i" * mdims, f.read(4 * mdims))
        if verbose:
            print("         dv:", dv)
        if mdims == 1:
            mdims = 2
            dv = [1, dv[0]]
        tmp = ord(f.read(1))
        if tmp == 0:
            raise OctaveReaderError("MatrixFormatError")
        num_elems = reduce(lambda x, y: x * y, dv)
        if verbose:
            print("  num_elems:", num_elems)
        num_doubles = 2 * num_elems
        fortran_vec = unpack("d" * num_doubles, f.read(8 * num_doubles))
        fortran_vec_complex = [
            fortran_vec[i] + fortran_vec[i + 1] * 1j for i in range(0, num_doubles, 2)
        ]
        var = np.array(fortran_vec_complex).reshape(dv, order="F")
        if verbose:
            print("     matrix:", var.shape)
        if mdims == 2 and dv[0] == 1:
            var = var.flatten()
    else:
        nr = mdims
        nc = unpack("i", f.read(4))[0]
        tmp = ord(f.read(1))
        if tmp == 0:
            raise OctaveReaderError("MatrixFormatError")
        dv = (nr, nc)
        num_elems = nr * nc
        num_doubles = 2 * num_elems
        if verbose:
            print("  num_elems:", num_elems)
        fortran_vec = unpack("d" * num_elems, f.read(8 * num_doubles))
        fortran_vec_complex = [
            fortran_vec[i] + fortran_vec[i + 1] * 1j for i in range(0, num_doubles, 2)
        ]
        var = np.array(fortran_vec_complex).reshape(dv, order="F")
        if verbose:
            print("     matrix:", var.shape)
    return var


def read_scalar_struct_var(f, verbose):
    """ Reads one scalar structure variable from Octave binary file and returns
    it as a dictionnary.

    See: octave_struct::load_binary from libinterp/octave-value/ov-struct.cc
    from GNU Octave source files

    """

    length = unpack("i", f.read(4))[0]
    if verbose:
        print("     length:", length)
    if length == 0:
        raise OctaveReaderError("ScalarStructFormatError")
    if length < 0:
        # ov-struct.cc says we has explicit dimensions in that case
        mdims = -length
        dv = unpack("i" * mdims, f.read(4 * mdims))
        if verbose:
            print("         dv:", dv)
        length = unpack("i", f.read(4))[0]
    else:
        dv = (1, 1)

    var = dict()
    if length > 0:
        for j in range(length):
            (name, varvar) = read_var(f, verbose)
            var[name] = varvar
    return var


def read_string_var(f, verbose):
    """ Reads one sq_string variable from Octave binary file

    See: octave_char_matrix_str::load_binary from octinterp/octave-value/ov-str-mat.cc
    from GNU Octave source files

    """

    elements = unpack("i", f.read(4))[0]
    if verbose:
        print("   elements:", elements)
    if elements == 0:
        raise OctaveReaderError("StringFormatError")
    if elements < 0:
        mdims = -elements
        dv = unpack("i" * mdims, f.read(4 * mdims))
        if verbose:
            print("         dv:", dv)
        num_elems = reduce(lambda x, y: x * y, dv)
        if verbose:
            print("  num_elems:", num_elems)
        fortran_vec = unpack("c" * num_elems, f.read(num_elems))
        var = np.array(fortran_vec).reshape(dv, order="F")
        if mdims == 2 and (dv[0] == 1 or dv[1] == 1):
            var = var.flatten().tobytes().decode("ascii")
        if verbose:
            print("     string:", var)
    else:
        for i in range(elements):
            length = unpack("i", f.read(4))[0]
            if length == 0:
                raise OctaveReaderError("StringFormatError")
            fortran_vec = f.read(length)
            var[i] = fortran_vec.decode("ascii")
    return var


def read_cell_var(f, verbose):
    """ Reads one Cell variable from Octave binary file and returns it as a list

    See: octave_cell::load_binary from octinterp/octave-value/ov-cell.cc
    from GNU Octave source files

    """

    mdims = unpack("i", f.read(4))[0]
    if mdims >= 0:
        raise OctaveReaderError("CellFormatError")
    if verbose:
        print("      mdims:", mdims)
    mdims = -mdims
    if mdims > 0:
        dv = unpack("i" * mdims, f.read(4 * mdims))
        if verbose:
            print("         dv:", dv)
        num_elems = reduce(lambda x, y: x * y, dv)
    else:
        num_elems = 0
    var = list()
    for i in range(num_elems):
        (name, varvar) = read_var(f, verbose)
        if name != "<cell-element>":
            raise OctaveReaderError("CellFormatError")
        var.append(varvar)
    return var


def read_var(f, verbose):
    """ Reads one variable from Octave binary file

    Format is described in libinterp/corefcn/ls-oct-binary.cc
    recalled below.

    Data (one set for each item):
    ============================

    object               type            bytes
    ------               ----            -----
    name_length          integer             4

    name                 string    name_length

    doc_length           integer             4

    doc                  string     doc_length

    global flag          integer             1

    data type            char                1

    In general "data type" is 255, and in that case the next arguments
    in the data set are

    object               type            bytes
    ------               ----            -----
    type_length          integer             4

    type                 string    type_length

    The string "type" is then used with octave_value_typeinfo::lookup_type
    to create an octave_value of the correct type. The specific load/save
    function is then called.

    For backward compatiablity "data type" can also be a value between 1
    and 7, where this defines a hardcoded octave_value of the type

    data type                  octave_value
    ---------                  ------------
    1                          scalar
    2                          matrix
    3                          complex scalar
    4                          complex matrix
    5                          string   (old style storage)
    6                          range
    7                          string

    Except for "data type" equal 5 that requires special treatment, these
    old style "data type" value also cause the specific load/save functions
    to be called. FILENAME is used for error messages.
    """

    name_length = unpack("i", f.read(4))[0]
    name = f.read(name_length).decode("ascii")
    if verbose:
        print('* Variable "' + name + '":')
    doc_length = unpack("i", f.read(4))[0]
    doc = f.read(doc_length).decode("ascii")
    if verbose:
        print("        doc:", doc)
    global_flag = ord(f.read(1))
    if verbose:
        print("     global:", (global_flag == 1))
    data_type = ord(f.read(1))
    if data_type == 1:
        data_type = "scalar"
    elif data_type == 2:
        data_type = "matrix"
    elif data_type == 3:
        data_type = "complex scalar"
    elif data_type == 4:
        data_type = "complex matrix"
    elif data_type == 5:
        data_type = "string"
    elif data_type == 6:
        data_type = "range"
    elif data_type == 7:
        data_type = "string"
    elif data_type == 255:
        data_type_length = unpack("i", f.read(4))[0]
        data_type = f.read(data_type_length).decode("ascii")
    else:
        raise OctaveReaderError("UnknownDataType")
    if verbose:
        print("  data_type:", data_type)
    if data_type == "scalar":
        var = read_scalar_var(f, verbose)
    elif data_type == "matrix":
        var = read_matrix_var(f, verbose)
    elif data_type == "complex matrix":
        var = read_complex_matrix_var(f, verbose)
    elif data_type == "scalar struct":
        var = read_scalar_struct_var(f, verbose)
    elif data_type == "string" or data_type == "sq_string" or data_type == "dq_string":
        # sq_string is single-quoted string
        # dq_string is double-quoted string
        # both are charMatrix
        var = read_string_var(f, verbose)
    elif data_type == "cell":
        var = read_cell_var(f, verbose)
    else:
        raise OctaveReaderError("NotImplemented: data_type =" + str(data_type))

    return (name, var)


def read_octave_binary(path, verbose=False):
    """ Reads an Octave binary file. Returns a dictionnary containing all the variables. """
    data = dict()
    with open(path, "rb") as f:
        _ = read_header(f, verbose)
        done = False
        while not done:
            try:
                name, var = read_var(f, verbose)
                data[name] = var
            except OctaveReaderError:
                done = True
                raise
            except error:
                done = True

    return data


def read_OctMI_session(sessionName, verbose=True, veryVerbose=False):
    filename = sessionName + "_MIstate.octave"
    if verbose:
        print("Loading saved MI session from file `" + filename + "'")
    MI_session = read_octave_binary(filename, veryVerbose)["MI_session"]
    Variables = dict()
    Variables["startTime"] = MI_session["startTime"]
    for varname in MI_session["Variables"]:
        keyname = varname + "_array"
        Variables[varname] = MI_session[keyname]
    time_fmt = "%A %e %B %Y - %H:%M:%S"
    if "t_array" in MI_session:
        Variables["t"] = MI_session["t_array"]
        try:
            Nelem = len(Variables["t"])
        except Exception:
            Nelem = 1
            pass
        if Nelem > 1:
            lt_start = localtime(Variables["t"][0])
            lt_end = localtime(Variables["t"][Nelem - 1])
        elif Nelem == 1:
            lt_start = localtime(Variables["t"])
            lt_end = lt_start
        else:
            lt_start = localtime(Variables["startTime"])
            lt_end = lt_start
    if verbose:
        string = strftime(time_fmt, lt_start)
        cprint.blue("** Start date: " + string)
        if Nelem > 0:
            string = strftime(time_fmt, lt_end)
            cprint.blue("**   End date: " + string)
        else:
            cprint.red("No logged variables")

    return Variables


if __name__ == "__main__":
    data_side = read_OctMI_session("W3E8_50Hz_side_1")
    data = read_OctMI_session("W3E8_3")
    dummy = read_OctMI_session("Dummy")



"""

Python wrapper for the NI System Configuration library

"""

from pprint import pprint
from ._lib.constants import *
from .resource import NISysCfgHardwareEnumerator, NISysCfgResourceAttributeError

preferred_experts = ["daqmx", "scope", "ni-visa", "ni-488.2"]


def find_resources():
    for res in NISysCfgHardwareEnumerator():
        try:
            attrs = [
                "VendorName",
                "ProductName",
                "SerialNumber",
                "ConnectsToLinkName",
                "PxiPciBusNumber",
                "PxiPciDeviceNumber",
                "PxiPciFunctionNumber",
                "GpibPrimaryAddress",
                "GpibSecondaryAddress",
                "ProvidesBusType",
                "ProvidesLinkName",
            ]
            infos = dict()
            for attr in attrs:
                try:
                    val = getattr(res, attr)
                    infos[attr] = val
                except NISysCfgResourceAttributeError as e:
                    if e.status != NISysCfgStatus.PropDoesNotExist:  # noqa: F405
                        raise
            expert_infos = res.indexed_properties()
            found = False
            for preferred in preferred_experts:
                for e in expert_infos:
                    if e["ExpertName"] == preferred:
                        infos[preferred] = e
                        found = True
                        break
                if found:
                    break

        finally:
            res.close()
        yield infos


def print_resources():
    for infos in find_resources():
        try:
            titre = infos["VendorName"] + " " + infos["ProductName"]
        except KeyError:
            continue
        print(titre)
        print("-" * len(titre))
        pprint(infos)


def daqmx_devices():
    return [
        info["daqmx"]["ExpertUserAlias"] for info in find_resources() if "daqmx" in info
    ]


def scope_devices():
    return [
        info["scope"]["ExpertUserAlias"] for info in find_resources() if "scope" in info
    ]



"""IDS Camera module (:mod:`pymanip.video.ids`)
===============================================

This module implements the :class:`pymanip.video.ids.IDS_Camera` class
using the third-party :mod:`pyueye` module.

.. autoclass:: IDS_Camera
   :members:
   :private-members:
   :show-inheritance:

"""

from datetime import datetime
import asyncio
import ctypes

import numpy as np
from pyueye import ueye
from pymanip.video import MetadataArray, Camera, CameraTimeout


class IDS_Camera(Camera):
    """Concrete implementation for IDS Camera.
    """

    def __init__(self, cam_num=0):
        self.name = f"IDS Camera {cam_num:}"
        self.hCam = ueye.HIDS(cam_num)

        self.sInfo = ueye.SENSORINFO()
        self.cInfo = ueye.CAMINFO()
        self.pcImageMemory = ueye.c_mem_p()
        self.MemID = ueye.int()
        self.rectAOI = ueye.IS_RECT()
        self.pitch = ueye.INT()
        self.nBitsPerPixel = ueye.INT(
            24
        )  # 24: bits per pixel for color mode; take 8 bits per pixel for monochrome
        self.channels = (
            3  # 3: channels for color mode(RGB); take 1 channel for monochrome
        )
        self.m_nColorMode = ueye.INT()  # Y8/RGB16/RGB24/REG32
        self.bytes_per_pixel = int(self.nBitsPerPixel / 8)

        # Starts the driver and establishes the connection to the camera
        nRet = ueye.is_InitCamera(self.hCam, None)
        if nRet != ueye.IS_SUCCESS:
            raise RuntimeError(f"is_InitCamera ERROR {nRet:}")

        # Reads out the data hard-coded in the non-volatile camera memory and writes it to the data structure that cInfo points to
        nRet = ueye.is_GetCameraInfo(self.hCam, self.cInfo)
        if nRet != ueye.IS_SUCCESS:
            raise RuntimeError("is_GetCameraInfo ERROR")

        # You can query additional information about the sensor type used in the camera
        nRet = ueye.is_GetSensorInfo(self.hCam, self.sInfo)
        if nRet != ueye.IS_SUCCESS:
            raise RuntimeError("is_GetSensorInfo ERROR")

        nRet = ueye.is_ResetToDefault(self.hCam)
        if nRet != ueye.IS_SUCCESS:
            raise RuntimeError("is_ResetToDefault ERROR")

        # Set display mode to DIB
        nRet = ueye.is_SetDisplayMode(self.hCam, ueye.IS_SET_DM_DIB)

        # Set the right color mode
        if (
            int.from_bytes(self.sInfo.nColorMode.value, byteorder="big")
            == ueye.IS_COLORMODE_BAYER
        ):
            # setup the color depth to the current windows setting
            ueye.is_GetColorDepth(self.hCam, self.nBitsPerPixel, self.m_nColorMode)
            self.bytes_per_pixel = int(self.nBitsPerPixel / 8)
            print("IS_COLORMODE_BAYER: ")
            print("\tm_nColorMode: \t\t", self.m_nColorMode)
            print("\tnBitsPerPixel: \t\t", self.nBitsPerPixel)
            print("\tbytes_per_pixel: \t\t", self.bytes_per_pixel)
            print()

        elif (
            int.from_bytes(self.sInfo.nColorMode.value, byteorder="big")
            == ueye.IS_COLORMODE_CBYCRY
        ):
            # for color camera models use RGB32 mode
            self.m_nColorMode = ueye.IS_CM_BGRA8_PACKED
            self.nBitsPerPixel = ueye.INT(32)
            self.bytes_per_pixel = int(self.nBitsPerPixel / 8)
            print("IS_COLORMODE_CBYCRY: ")
            print("\tm_nColorMode: \t\t", self.m_nColorMode)
            print("\tnBitsPerPixel: \t\t", self.nBitsPerPixel)
            print("\tbytes_per_pixel: \t\t", self.bytes_per_pixel)
            print()

        elif (
            int.from_bytes(self.sInfo.nColorMode.value, byteorder="big")
            == ueye.IS_COLORMODE_MONOCHROME
        ):
            # for color camera models use RGB32 mode
            self.m_nColorMode = ueye.IS_CM_MONO8
            self.nBitsPerPixel = ueye.INT(8)
            self.bytes_per_pixel = int(self.nBitsPerPixel / 8)
            print("IS_COLORMODE_MONOCHROME: ")
            print("\tm_nColorMode: \t\t", self.m_nColorMode)
            print("\tnBitsPerPixel: \t\t", self.nBitsPerPixel)
            print("\tbytes_per_pixel: \t\t", self.bytes_per_pixel)
            print()

        else:
            # for monochrome camera models use Y8 mode
            self.m_nColorMode = ueye.IS_CM_MONO8
            self.nBitsPerPixel = ueye.INT(8)
            self.bytes_per_pixel = int(self.nBitsPerPixel / 8)
            # print("else")

        # Can be used to set the size and position of an "area of interest"(AOI) within an image
        nRet = ueye.is_AOI(
            self.hCam,
            ueye.IS_AOI_IMAGE_GET_AOI,
            self.rectAOI,
            ueye.sizeof(self.rectAOI),
        )
        if nRet != ueye.IS_SUCCESS:
            print("is_AOI ERROR")

        self.width = self.rectAOI.s32Width
        self.height = self.rectAOI.s32Height

        # Prints out some information about the camera and the sensor
        print("Camera model:\t\t", self.sInfo.strSensorName.decode("utf-8"))
        print("Camera serial no.:\t", self.cInfo.SerNo.decode("utf-8"))
        print("Maximum image width:\t", self.width)
        print("Maximum image height:\t", self.height)
        print()

        nRet = ueye.is_AllocImageMem(
            self.hCam,
            self.width,
            self.height,
            self.nBitsPerPixel,
            self.pcImageMemory,
            self.MemID,
        )
        if nRet != ueye.IS_SUCCESS:
            print("is_AllocImageMem ERROR")
        else:
            # Makes the specified image memory the active memory
            nRet = ueye.is_SetImageMem(self.hCam, self.pcImageMemory, self.MemID)
            if nRet != ueye.IS_SUCCESS:
                print("is_SetImageMem ERROR")
            else:
                # Set the desired color mode
                nRet = ueye.is_SetColorMode(self.hCam, self.m_nColorMode)

    def close(self):
        ueye.is_ExitCamera(self.hCam)

    def __exit__(self, type_, value, cb):
        """Context manager exit method
        """
        super(IDS_Camera, self).__exit__(type_, value, cb)
        self.close()

    def acquisition_oneshot(self, timeout_ms=1000):
        nRet = ueye.is_EnableEvent(self.hCam, ueye.IS_SET_EVENT_FRAME)
        if nRet != ueye.IS_SUCCESS:
            raise RuntimeError("is_EnableEvent ERROR")

        nRet = ueye.is_FreezeVideo(self.hCam, ueye.IS_DONT_WAIT)
        if nRet != ueye.IS_SUCCESS:
            raise RuntimeError("is_CaptureVideo ERROR")
        nRet = ueye.is_WaitEvent(self.hCam, ueye.IS_SET_EVENT_FRAME, timeout_ms)
        if nRet != ueye.IS_SUCCESS:
            raise RuntimeError("is_WaitEvent ERROR")

        nRet = ueye.is_InquireImageMem(
            self.hCam,
            self.pcImageMemory,
            self.MemID,
            self.width,
            self.height,
            self.nBitsPerPixel,
            self.pitch,
        )
        if nRet != ueye.IS_SUCCESS:
            raise RuntimeError("is_InquireImageMem ERROR")
        array = ueye.get_data(
            self.pcImageMemory,
            self.width,
            self.height,
            self.nBitsPerPixel,
            self.pitch,
            copy=True,
        )

        nRet = ueye.is_DisableEvent(self.hCam, ueye.IS_SET_EVENT_FRAME)
        if nRet != ueye.IS_SUCCESS:
            raise RuntimeError("is_DisableEvent ERROR")

        return array.reshape((self.height.value, self.width.value))

    async def acquisition_async(
        self,
        num=np.inf,
        timeout=1000,
        raw=False,
        initialising_cams=None,
        raise_on_timeout=True,
    ):
        """Concrete implementation
        """
        loop = asyncio.get_event_loop()

        nRet = ueye.is_CaptureVideo(self.hCam, ueye.IS_DONT_WAIT)
        if nRet != ueye.IS_SUCCESS:
            raise RuntimeError("is_CaptureVideo ERROR")

        try:
            nRet = ueye.is_InquireImageMem(
                self.hCam,
                self.pcImageMemory,
                self.MemID,
                self.width,
                self.height,
                self.nBitsPerPixel,
                self.pitch,
            )
            image_info = ueye.UEYEIMAGEINFO()
            if nRet != ueye.IS_SUCCESS:
                raise RuntimeError("is_InquireImageMem ERROR")

            count = 0
            while count < num:
                nRet = ueye.is_EnableEvent(self.hCam, ueye.IS_SET_EVENT_FRAME)
                if nRet != ueye.IS_SUCCESS:
                    raise RuntimeError("is_EnableEvent ERROR")
                nRet = await loop.run_in_executor(
                    None, ueye.is_WaitEvent, self.hCam, ueye.IS_SET_EVENT_FRAME, timeout
                )
                if nRet == ueye.IS_TIMED_OUT:
                    if raise_on_timeout:
                        raise RuntimeError("Timeout")
                    else:
                        stop_signal = yield None
                        if stop_signal:
                            break
                        else:
                            continue
                elif nRet != ueye.IS_SUCCESS:
                    raise RuntimeError("is_WaitEvent ERROR")
                array = ueye.get_data(
                    self.pcImageMemory,
                    self.width,
                    self.height,
                    self.nBitsPerPixel,
                    self.pitch,
                    copy=False,
                )

                nRet = ueye.is_GetImageInfo(
                    self.hCam, self.MemID, image_info, ctypes.sizeof(image_info)
                )
                if nRet != ueye.IS_SUCCESS:
                    raise RuntimeError("is_GetImageInfo ERROR")

                count = count + 1
                ts = datetime(
                    image_info.TimestampSystem.wYear.value,
                    image_info.TimestampSystem.wMonth.value,
                    image_info.TimestampSystem.wDay.value,
                    image_info.TimestampSystem.wHour.value,
                    image_info.TimestampSystem.wMinute.value,
                    image_info.TimestampSystem.wSecond.value,
                    image_info.TimestampSystem.wMilliseconds * 1000,
                )
                stop_signal = yield MetadataArray(
                    array.reshape((self.height.value, self.width.value)),
                    metadata={"counter": count, "timestamp": ts.timestamp()},
                )
                if stop_signal:
                    break

        finally:
            nRet = ueye.is_StopLiveVideo(self.hCam, ueye.IS_DONT_WAIT)
            if nRet != ueye.IS_SUCCESS:
                raise RuntimeError("is_StopLiveVideo ERROR")
            nRet = ueye.is_DisableEvent(self.hCam, ueye.IS_SET_EVENT_FRAME)
            if nRet != ueye.IS_SUCCESS:
                raise RuntimeError("is_DisableEvent ERROR")
        if stop_signal:
            yield True

    def possible_pixelclock(self):
        """Query the possible values for pixelclock (in MHz)
        """

        nPixelclocks = ueye.UINT()
        nRet = ueye.is_PixelClock(
            self.hCam,
            ueye.IS_PIXELCLOCK_CMD_GET_NUMBER,
            nPixelclocks,
            ctypes.sizeof(nPixelclocks),
        )
        if nRet != ueye.IS_SUCCESS:
            raise RuntimeError("IS_PIXELCLOCK_CMD_GET_NUMBER failed")
        if nPixelclocks.value == 0:
            return []
        pixelclock_list = (ueye.UINT * nPixelclocks.value)()
        nRet = ueye.is_PixelClock(
            self.hCam,
            ueye.IS_PIXELCLOCK_CMD_GET_LIST,
            pixelclock_list,
            ctypes.sizeof(pixelclock_list),
        )
        if nRet != ueye.IS_SUCCESS:
            raise RuntimeError("IS_PIXELCLOCK_CMD_GET_LIST failed")
        return [pc.value for pc in pixelclock_list]

    def set_pixelclock(self, pixelclock):
        """Set the pixelclock (in MHz)
        """
        pc = ueye.UINT(int(pixelclock))
        nRet = ueye.is_PixelClock(
            self.hCam, ueye.IS_PIXELCLOCK_CMD_SET, pc, ctypes.sizeof(pc)
        )
        if nRet != ueye.IS_SUCCESS:
            raise RuntimeError("IS_PIXELCLOCK_CMD_SET failed")
        return pc.value

    def current_pixelclock(self):
        """Queries the current pixelclock (in MHz)
        """
        pc = ueye.UINT()
        nRet = ueye.is_PixelClock(
            self.hCam, ueye.IS_PIXELCLOCK_CMD_GET, pc, ctypes.sizeof(pc)
        )
        if nRet != ueye.IS_SUCCESS:
            raise RuntimeError("IS_PIXELCLOCK_CMD_GET failed")
        return pc.value

    def possible_exposure_time(self):
        """Query the min, max and increment in ms
        """
        min_ms = ctypes.c_double()
        max_ms = ctypes.c_double()
        inc_ms = ctypes.c_double()
        nRet = ueye.is_Exposure(
            self.hCam,
            ueye.IS_EXPOSURE_CMD_GET_EXPOSURE_RANGE_MIN,
            min_ms,
            ctypes.sizeof(min_ms),
        )
        if nRet != ueye.IS_SUCCESS:
            raise RuntimeError("IS_EXPOSURE_CMD_GET_EXPOSURE_RANGE_MIN failed")
        nRet = ueye.is_Exposure(
            self.hCam,
            ueye.IS_EXPOSURE_CMD_GET_EXPOSURE_RANGE_MAX,
            max_ms,
            ctypes.sizeof(max_ms),
        )
        if nRet != ueye.IS_SUCCESS:
            raise RuntimeError("IS_EXPOSURE_CMD_GET_EXPOSURE_RANGE_MAX failed")
        nRet = ueye.is_Exposure(
            self.hCam,
            ueye.IS_EXPOSURE_CMD_GET_EXPOSURE_RANGE_INC,
            inc_ms,
            ctypes.sizeof(inc_ms),
        )
        if nRet != ueye.IS_SUCCESS:
            raise RuntimeError("IS_EXPOSURE_CMD_GET_EXPOSURE_RANGE_INC failed")
        return min_ms.value, max_ms.value, inc_ms.value

    def current_exposure_time(self):
        """Query the current exposure time in ms
        """
        exposure_ms = ctypes.c_double()
        nRet = ueye.is_Exposure(
            self.hCam,
            ueye.IS_EXPOSURE_CMD_GET_EXPOSURE,
            exposure_ms,
            ctypes.sizeof(exposure_ms),
        )
        if nRet != ueye.IS_SUCCESS:
            raise RuntimeError("IS_EXPOSURE_CMD_GET_EXPOSURE failed")

        return exposure_ms.value

    def set_exposure_time(self, exposure_ms):
        """Sets exposure time in ms.
        If 0 is passed, the exposure time is set to the maximum value of 1/frame rate.
        """
        exposure_ms_double = ctypes.c_double(exposure_ms)
        nRet = ueye.is_Exposure(
            self.hCam,
            ueye.IS_EXPOSURE_CMD_SET_EXPOSURE,
            exposure_ms_double,
            ctypes.sizeof(exposure_ms_double),
        )
        if nRet != ueye.IS_SUCCESS:
            raise RuntimeError("IS_EXPOSURE_CMD_SET_EXPOSURE failed")
        actual = exposure_ms_double.value
        if actual != exposure_ms:
            print("Warning: actual value of exposure time is", actual, "ms")

    def set_frame_rate(self, framerate_fps):
        """Sets the framerate in frames per seconds
        """
        newFPS = ctypes.c_double()
        nRet = ueye.is_SetFrameRate(self.hCam, float(framerate_fps), newFPS)
        if nRet != ueye.IS_SUCCESS:
            raise RuntimeError("SetFrameRate failed")
        if newFPS.value != framerate_fps:
            print("Warning actual framerate is", newFPS.value)
        return newFPS.value

    def current_frame_rate(self):
        """Queries the current framerate
        """
        fps = ctypes.c_double()
        nRet = ueye.is_GetFrameRate(self.hCam, fps)
        if nRet != ueye.IS_SUCCESS:
            raise RuntimeError("GetFrameRate failed")
        return fps.value



"""Asynchronous extension of :class:`fluidlab.interfaces.VISAInterface` (:mod:`pymanip.interfaces.aiovisa`)
===========================================================================================================

This module defines :class:`AsyncVISAInterface` as a subclass of :class:`fluidlab.interfaces.visa_inter.VISAInterface` and
:class:`pymanip.interface.aiointer.AsyncQueryInterface`.

.. autoclass:: AsyncVISAInterface
   :members:
   :private-members:

"""

import asyncio
import fluidlab.interfaces.visa_inter as fl_visa
from pymanip.interfaces.aiointer import AsyncQueryInterface

VISALock = asyncio.Lock()


class AsyncVISAInterface(fl_visa.VISAInterface, AsyncQueryInterface):
    """This class is an asynchronous extension of :class:`fluidlab.interfaces.visa_inter.VISAInterface`.
    The parameters are the same as those of the :class:`fluidlab.interfaces.visa_inter.VISAInterface` class.

    """

    def __init__(self, resource_name, backend=None):
        """Constructor method
        """
        super().__init__(resource_name, backend)
        # Override interface lock with global VISA lock because VISA library is not
        # thread-safe so we don't want concurrent calls to VISA lib.
        self.lock = VISALock

    async def await_for_srq(self, timeout=None):
        """This co-routine method acquires the interface lock and run wait_for_srq in an
        executor.
        """
        async with self.lock:
            await self.loop.run_in_executor(None, self.wait_for_srq, timeout)



"""Asynchronous extension of :class:`fluidlab.interfaces.QueryInterface` (:mod:`pymanip.interfaces.aiointer`)
=============================================================================================================

This module defines :class:`AsyncQueryInterface` as the default subclass for :class:`fluidlab.interfaces.QueryInterface`.
The default implementation simply runs the methods of :class:`QueryInterface` into an executor, therefore in a separate thread. Because the parent class is probably not thread-safe, each call is protected by a lock to prevent concurrent calls to the instrument. However, the main thread is released and other tasks can be run on other instruments.

The methods that are defined in this way are :meth:`__aenter__`, :meth:`__aexit__`, :meth:`_aread`, :meth:`_awrite`, :meth:`_aquery` and :meth:`await_for_srq`. The higher lever co-routine methods :meth:`aread`, :meth:`awrite` and :meth:`aquery` simply checks that the interface is opened, and then awaits the low-level method, in a similar fashion as in the :class:`QueryInterface` class.

Therefore, concrete subclass such as :class:`pymanip.interfaces.aioserial.AsyncSerialInterface`, or :class:`pymanip.interfaces.aiovisa.AsyncVISAInterface` only have to override the low-level co-routine methods (if necessary).

.. autoclass:: AsyncQueryInterface
   :members:
   :private-members:

"""

import warnings
import asyncio
import fluidlab.interfaces as flinter
import functools
from time import monotonic


class AsyncQueryInterface(flinter.QueryInterface):
    """This class represents an asynchronous Query interface. It is a subclass of the synchronous
    QueryInterface defined in FluidLab. The input parameters are those of QueryInterface.

    Concrete subclasses may replace the :attr:`lock` attribute, and replace it by a global board lock if necessary.
    """

    def __init__(self):
        """Constructor method
        """

        super().__init__()
        self.loop = asyncio.get_event_loop()
        self.lock = (
            asyncio.Lock()
        )  # dummy lock since it is not shared by another interface,
        # but it prevents the same device to concurrently talk

    def __str__(self):
        return "Async" + super().__str__()

    def __repr__(self):
        return str(self)

    async def __aenter__(self):
        """Asynchronous context manager enter method
        """
        async with self.lock:
            return await self.loop.run_in_executor(None, self.__enter__)

    async def __aexit__(self, type_, value, tb):
        """Asynchronous context manager exit method
        """
        async with self.lock:
            await self.loop.run_in_executor(None, self.__exit__, type_, value, tb)

    async def _awrite(self, *args, **kwargs):
        """Low-level co-routine to send data to the instrument. This method does not check whether the interface
        is opened.
        In this basic class, it simply acquires the interface lock and runs the :meth:`_write` method in an executor.
        """
        async with self.lock:
            await self.loop.run_in_executor(
                None, self._write, *args
            )  # must be rewritten if necessary in concrete class

    async def _aread(self, *args, **kwargs):
        """Low-level co-routine to read data from the instrument. This method does not check whether the interface
        is opened.
        In this basic class, it simply acquires the interface lock and runs the :meth:`_read` method in an executor.
        """
        async with self.lock:
            data = await self.loop.run_in_executor(None, self._read, *args)
        return data

    async def _aquery(self, command, time_delay=0.1, **kwargs):
        """Low-level co-routine to write/read a query. This method does not check whether the interface is opened.
        There are two cases:

        - if the QueryInterface has a :meth:`_query` method, then the interface lock is acquired and the :meth:`_query` method is run in an executor;

        - otherwise, the :meth:`awrite` and :meth:`aread` co-routine methods are used.

        """

        if hasattr(self, "_query"):
            query_func = functools.partial(
                self.query, command, time_delay=time_delay, **kwargs
            )
            async with self.lock:
                return await self.loop.run_in_executor(None, query_func)
        else:
            if hasattr(self, "timeout"):
                timeout = self.timeout
            elif "timeout" in kwargs:
                timeout = kwargs["timeout"]
            else:
                timeout = 60.0

            if timeout < 10:
                timeout = 10.0

            await self.awrite(command)
            start = monotonic()
            while True:
                await asyncio.sleep(time_delay)
                data = await self.aread(**kwargs)
                if data:
                    return data
                elif monotonic() - start > timeout:
                    print("timeout")
                    break
                await asyncio.sleep(9 * time_delay)

    async def awrite(self, *args, **kwargs):
        """This co-routine method writes data to the instrument. The parameters are identical to the
        :meth:`write` method.
        """
        if not self.opened:
            warnings.warn(
                "write() called on non-opened interface.", flinter.InterfaceWarning
            )
            self.open()
        await self._awrite(*args, **kwargs)

    async def aread(self, *args, **kwargs):
        """This co-routine method reads data from the instrument. The parameters are identical to
        the :meth:`read` method.
        """
        if not self.opened:
            warnings.warn(
                "read() called on non-opened interface.", flinter.InterfaceWarning
            )
            self.open()
        return await self._aread(*args, **kwargs)

    async def aquery(self, command, time_delay=0.1, **kwargs):
        """This co-routine method queries the instrument. The parameters are identical to those of
        the :meth:`query` method.
        """
        if not self.opened:
            warnings.warn(
                "query() called on non-opened interface.", flinter.InterfaceWarning
            )
            self.open()
        return await self._aquery(command, time_delay=time_delay, **kwargs)



import warnings

from .session import NISysCfgSession
from ._lib.session import (
    NISysCfgFindHardware,
    NISysCfgCloseHandle,
    NISysCfgNextResource,
)
from ._lib.properties import (
    NISysCfgGetResourceProperty,
    NISysCfgGetResourceIndexedProperty,
)
from ._lib.constants import NISysCfgResourceProperty, NISysCfgStatus


class NISysCfgResourceAttributeError(AttributeError):
    def __init__(self, description, status):
        self.status = status
        super(NISysCfgResourceAttributeError, self).__init__(description)


class NISysCfgResource:
    def __init__(self, resourceHandle):
        self.resourceHandle = resourceHandle

    def close(self):
        if self.resourceHandle:
            status = NISysCfgCloseHandle(self.resourceHandle)
            if status != NISysCfgStatus.OK:
                raise RuntimeError(
                    f"NISysCfgCloseHandle failed. Status: {str(status):}."
                )
            self.resourceHandle = None
        else:
            warnings.warn("close was called twice on NISysCfgResource!")

    def __del__(self):
        if self.resourceHandle is not None:
            warnings.warn("NISysCfgResource was left open!", RuntimeWarning)
            self.close()

    def __getattr__(self, key):
        status, val = NISysCfgGetResourceProperty(
            self.resourceHandle, getattr(NISysCfgResourceProperty, key)
        )
        if status == NISysCfgStatus.OK:
            return val
        raise NISysCfgResourceAttributeError(
            f"NISysCfgGetResourceProperty returned {str(status):}.", status
        )

    def indexed_properties(self):
        return NISysCfgGetResourceIndexedProperty(self.resourceHandle)


class _NISysCfgHardwareEnumerator:
    def __init__(self, sesn):
        self.sesn = sesn
        self.resourceEnumHandle = None

    def __enter__(self):
        status, resourceEnumHandle = NISysCfgFindHardware(self.sesn.sessionHandle)
        if status != NISysCfgStatus.OK:
            raise RuntimeError(f"NISysCfgFindHardware failed. Status: {str(status):}.")
        self.resourceEnumHandle = resourceEnumHandle
        return self

    def __exit__(self, type_, value, cb):
        if self.resourceEnumHandle:
            status = NISysCfgCloseHandle(self.resourceEnumHandle)
            if status != NISysCfgStatus.OK:
                raise RuntimeError(
                    f"NISysCfgCloseHandle failed. Status: {str(status):}."
                )

    def next_resource(self):
        status, resourceHandle = NISysCfgNextResource(
            self.sesn.sessionHandle, self.resourceEnumHandle
        )
        if status == NISysCfgStatus.OK and resourceHandle:
            return NISysCfgResource(resourceHandle)
        else:
            return None


def NISysCfgHardwareEnumerator():
    with NISysCfgSession() as sesn, _NISysCfgHardwareEnumerator(sesn) as enum:

        while True:
            res = enum.next_resource()
            if res:
                yield res
            else:
                break



"""Asynchronous TDK Lambda extension (:mod:`pymanip.aioinstruments.tdk_lambda`)
===============================================================================

.. autoclass:: AsyncTdkLambda
   :members:
   :private-members:

"""

import fluidlab.instruments.powersupply.tdk_lambda as fl_tdk
from pymanip.aioinstruments.aiofeatures import AsyncFloatValue, AsyncBoolValue
from pymanip.aioinstruments.aioiec60488 import AsyncIEC60488


class AsyncTdkLambda(AsyncIEC60488, fl_tdk.TdkLambda):
    pass


afeatures = [
    AsyncFloatValue(
        "idc",
        doc="Get actual current/Set current setpoint",
        command_get="MEAS:CURR?",
        command_set=":CURRENT",
        check_instrument_value=False,
        pause_instrument=0.1,
    ),
    AsyncFloatValue(
        "vdc",
        doc="Get actual voltage/Set voltage setpoint",
        command_get="MEAS:VOLT?",
        command_set="SOUR:VOLT",
        check_instrument_value=False,
        pause_instrument=0.1,
    ),
    AsyncBoolValue(
        "onoff",
        doc="Toggle output ON/OFF",
        command_set="OUTPUT:STATE",
        check_instrument_value=False,
        pause_instrument=0.1,
    ),
]

AsyncTdkLambda._build_class_with_features(afeatures)



import os
import pytest
import pymanip.session as sess


def test_name_generator():
    """

    Test the name generator

    """

    name1 = sess.makeAcqName()
    name2 = sess.makeAcqName()
    assert name1 != name2


def test_log_addline(tmpdir):
    """

    Test the log_addline function

    """

    with sess.Session(os.path.join(tmpdir, "test_session"), ("a", "b", "c")) as sesn:
        for a in range(5):
            b = a ** 2
            c = 2 * b  # noqa: F841
            sesn.log_addline()

    with sess.SavedSession(os.path.join(tmpdir, "test_session")) as sesn:
        assert tuple(sesn.log_variable_list()) == ("a", "b", "c")
        assert sesn.has_log("a")
        assert sesn.has_log("b")
        assert sesn.has_log("c")
        assert (sesn["a"] == (0, 1, 2, 3, 4)).all()
        assert (sesn["b"] == (0, 1, 4, 9, 16)).all()
        assert (sesn["c"] == (0, 2, 8, 18, 32)).all()
        assert (sesn.log("a") == sesn["a"]).all()
        assert len(tuple(sesn.dataset_names())) == 0
        assert not sesn.has_parameter("toto")
        with pytest.raises(KeyError):
            sesn["toto"]


def test_cache(tmpdir):
    """

    Test the SavedSession cache system

    """

    # Create a dummy session
    with sess.Session(os.path.join(tmpdir, "test_session"), ("a", "b", "c")) as sesn:
        for a in range(5):
            b = a ** 2
            c = 2 * b  # noqa: F841
            sesn.log_addline()

    # Test the cache
    with sess.SavedSession(
        os.path.join(tmpdir, "test_session"), cache_location=tmpdir
    ) as sesn:
        assert len(tuple(sesn.cachedvars)) == 0
        assert not sesn.cached("toto")
        with pytest.raises(KeyError):
            sesn.cachedvalue("toto")
        toto = [10, 13, 15]  # noqa: F841
        sesn.cache("toto")

    sesn = None

    with sess.SavedSession(
        os.path.join(tmpdir, "test_session"), cache_location=tmpdir
    ) as sesn:
        assert tuple(sesn.cachedvars) == ("toto",)
        assert sesn.cached("toto")
        assert (sesn.cachedvalue("toto") == [10, 13, 15]).all()
        with pytest.raises(KeyError):
            sesn["toto"]



"""

Utility function for cameras and videos

"""

try:
    from pymanip.video.pco import PCO_Camera

    has_pco = True
except Exception:
    has_pco = False
try:
    from pymanip.video.avt import AVT_Camera

    has_avt = True
except Exception:
    has_avt = False
try:
    from pymanip.video.andor import Andor_Camera

    has_andor = True
except Exception:
    has_andor = False

try:
    from pymanip.video.ids import IDS_Camera

    has_ids = True
except Exception:
    has_ids = False

try:
    from pymanip.video.ximea import Ximea_Camera

    has_ximea = True
except Exception:
    has_ximea = False

try:
    from pymanip.video.photometrics import Photometrics_Camera
    has_photometrics = True
except Exception as e:
    has_photometrics = False
    photometrics_exc = e

from contextlib import ExitStack
import asyncio


def preview_pco(
    interface="all",
    board=0,
    backend="cv",
    slice=None,
    zoom=0.5,
    TriggerMode=None,
    exposure_ms=20,
    rotate=0,
    roi=None,
):
    if not has_pco:
        print("PCO bindings are not available.")
    else:
        with PCO_Camera(interface, board) as cam:
            if TriggerMode:
                cam.set_trigger_mode(TriggerMode)
            else:
                cam.set_trigger_mode("auto sequence")
            cam.set_delay_exposuretime(exposuretime=exposure_ms / 1000)
            if roi is None:
                res = cam.resolution
                roi = (1, 1, res[0], res[1])
            cam.set_roi(*roi)
            cam.preview(backend, slice, zoom, rotate)


def preview_avt(
    board=0,
    backend="cv",
    slice=None,
    zoom=0.5,
    TriggerMode=None,
    exposure_ms=10,
    rotate=0,
    roi=None,
):
    if not has_avt:
        print("Pymba is not available.")
    else:
        if backend == "cv":
            print("Press 's' to close window")
        if isinstance(board, list) and len(board) == 1:
            board = board[0]
        if isinstance(board, list):
            with ExitStack() as stack:
                cams = [stack.enter_context(AVT_Camera(b)) for b in board]
                if TriggerMode:
                    print("External trigger")
                    for c in cams:
                        c.set_trigger_mode(True)
                else:
                    print("Internal trigger")
                    for c in cams:
                        c.set_trigger_mode(False)
                if roi:
                    for c in cams:
                        c.set_roi(*roi)
                loop = asyncio.get_event_loop()
                loop.run_until_complete(
                    asyncio.gather(
                        *[
                            c.preview_async_cv(
                                slice, zoom, name="AVT " + str(c.num), rotate=rotate
                            )
                            for c in cams
                        ]
                    )
                )
                loop.close()
        else:
            with AVT_Camera(board) as cam:
                if TriggerMode:
                    print("External trigger")
                    cam.set_trigger_mode(True)
                else:
                    print("Internal trigger")
                    cam.set_trigger_mode(False)
                if roi:
                    cam.set_roi(*roi)
                cam.set_exposure_time(exposure_ms / 1000)
                cam.preview(backend, slice, zoom, rotate)


def preview_andor(
    num=0,
    backend="cv",
    slice=None,
    zoom=1.0,
    TriggerMode=None,
    exposure_ms=1,
    bitdepth=12,
    framerate=10.0,
    rotate=0,
):
    if not has_andor:
        print("Andor bindings are not available.")
    else:
        with Andor_Camera(num) as cam:
            cam.set_exposure_time(exposure_ms / 1000)
            cam.FrameRate.setValue(framerate)
            if bitdepth == 12:
                cam.PixelEncoding.setString("Mono12Packed")  # Mono12Packed Mono16
                # cam.BitDepth.setString('12 Bit')
                cam.SimplePreAmpGainControl.setString("11-bit (low noise)")
            elif bitdepth == 16:
                cam.PixelEncoding.setString("Mono16")
                # cam.BitDepth.setString('16 Bit')
                cam.SimplePreAmpGainControl.setString(
                    "16-bit (low noise & high well capacity)"
                )
            else:
                raise ValueError("Only 12-bits or 16-bits")
            cam.preview(backend, slice, zoom, rotate)


def preview_ids(
    num=0,
    backend="cv",
    slice=None,
    zoom=1.0,
    TriggerMode=None,
    exposure_ms=1,
    bitdepth=12,
    framerate=10.0,
    rotate=0,
):
    if not has_ids:
        print("IDS bindings (pyueyes) are not available")
    else:
        with IDS_Camera(num) as cam:
            cam.set_exposure_time(exposure_ms)
            cam.set_frame_rate(framerate)
            cam.preview(backend, slice, zoom, rotate)


def preview_ximea(
    num=None,
    backend="cv",
    slice=None,
    zoom=1.0,
    TriggerMode=None,
    exposure_ms=1.0,
    rotate=0,
    white_balance=False,
    roi=None,
):

    if not has_ximea:
        print("Ximea bindings are not available.")
    else:
        with Ximea_Camera(num) as cam:
            cam.set_trigger_mode(TriggerMode)
            cam.set_exposure_time(exposure_ms * 1e-3)
            if cam.cam.is_iscolor():
                cam.set_auto_white_balance(white_balance)
            if roi is not None:
                cam.set_roi(*roi)
            cam.preview(backend, slice, zoom, rotate)

def preview_photometrics(
    num=0,
    backend="cv",
    slice=None,
    zoom=1.0,
    TriggerMode=None,
    exposure_ms=1.0,
    bitdepth=12,
    rotate=0,
    roi=None,
):
    """Bitdepth is 8, 12 or 16. Then we choose readout_port "Speed" (1), "Sensitivity" (0) or "Dynamic Range" (2)
    """
    if not has_photometrics:
        print("Photometrics bindings are not available.")
        raise photometrics_exc
    else:
        if bitdepth == 8:
            readout_port = 1
        elif bitdepth == 12:
            readout_port = 0
        elif bitdepth == 16:
            readout_port = 2
        with Photometrics_Camera(num, readout_port) as cam:
            cam.set_exposure_time(exposure_ms * 1e-3)
            try:
                cam.set_trigger_mode(TriggerMode)
            except RuntimeError:
                pass
            cam.preview(backend, slice, zoom, rotate)


from importlib.resources import path

import numpy as np

from pymanip.legacy_session.octmi_binary import read_octave_binary, read_OctMI_session
from pymanip.legacy_session.octmi_dat import load_octmi_dat
import pymanip.legacy_session.example as example


def test_octave_binary():
    """

    Test reading from Octave binary file

    """

    with path(example, "essai.octave") as p:
        data = read_octave_binary(p)

    assert "A" in data
    assert "B" in data
    assert (data["A"] == np.array([1, 2, 3, 4])).all()
    assert (data["B"] == 2 * data["A"]).all()


def test_octmi_session():
    """

    Test reading from OctMI binary session file
    We need to use isclose() for the cosine test, because the exact
    result apparently depends on the platform (it would fail
    on travis, but pass on the Mac).

    """

    with path(example, "essai2_MIstate.octave") as p:
        dirpath = p.parent
        data = read_OctMI_session(str(dirpath / "essai2"))

    assert "startTime" in data
    assert "A" in data
    assert "B" in data
    assert "t" in data
    assert data["startTime"] == 1562334542.728428
    assert (data["A"] == np.arange(0.0, 5.1, 0.1)).all()
    assert np.isclose(data["B"], np.cos(data["A"])).all()
    assert data["t"].size == data["A"].size
    assert data["t"][0] > data["startTime"]


def test_octmi_dat():
    """

    Test reading from OctMI dat session file.
    We need to use np.isclose() because conversion to/from ascii may change
    one or two bits on the double precision numbers.

    """

    with path(example, "essai2_MI.dat") as p:
        dirpath = p.parent
        data = load_octmi_dat(str(dirpath / "essai2"))

    assert "Time" in data
    assert "A" in data
    assert "B" in data
    assert data["nval"] == 51
    assert np.isclose(data["A"], np.arange(0.0, 5.1, 0.1)).all()
    assert np.isclose(data["B"], np.cos(data["A"]), rtol=1e-4).all()



"""Ximea Camera module (:mod:`pymanip.video.ximea.camera`)
==========================================================

This module implements the :class:`pymanip.video.ximea.Ximea_Camera` using the
`Python module provided by Ximea <https://www.ximea.com/support/wiki/apis/Python_inst_win>`_.

.. autoclass:: Ximea_Camera
   :members:
   :private-members:
   :show-inheritance:

"""

import asyncio
import ctypes

import numpy as np
from pymanip.video import MetadataArray, Camera, CameraTimeout

from ximea import xiapi
from ximea.xidefs import XI_IMG_FORMAT, XI_TRG_SOURCE, XI_TRG_SELECTOR


class Ximea_Camera(Camera):
    """Concrete :class:`pymanip.video.Camera` class for Ximea camera.
    """

    color_order = "BGR"  # Ximea RGB mode is [blue][green][red] per the doc.

    def __init__(self, serial_number=None, pixelFormat=None):
        """Constructor method
        """
        # pour le moment on ouvre simplement la première caméra
        self.cam = xiapi.Camera()
        if serial_number is None:
            self.cam.open_device()
        else:
            self.cam.open_device_by_SN(serial_number)
        if pixelFormat is None:
            if self.cam.is_iscolor():
                pixelFormat = "XI_RGB24"
            else:
                pixelFormat = "XI_MONO8"
        elif pixelFormat not in XI_IMG_FORMAT:
            raise ValueError(
                f"Wrong pixelFormat. Possible values are {set(XI_IMG_FORMAT.keys()):}"
            )
        self.cam.set_imgdataformat(pixelFormat)
        print(f"Camera {self.name:} opened ({pixelFormat:})")

    def close(self):
        """Close connection to the camera
        """
        name = self.name
        self.cam.close_device()
        self.cam = None
        print(f"Connection to camera {name:} closed.")

    def __exit__(self, type_, value, cb):
        """Context manager exit method
        """
        super().__exit__(type_, value, cb)
        self.close()

    @property
    def name(self):
        """Camera name
        """
        return self.cam.get_device_name().decode("ascii")

    def set_exposure_time(self, seconds):
        """This method sets the exposure time for the camera.

        :param seconds: exposure in seconds.
        :type seconds: float
        """
        self.cam.set_exposure(int(seconds * 1e6))
        print("Exposure time set to", 1000 * seconds, "ms")

    def get_exposure_time(self):
        """This method gets the exposure time in seconds for the camera.
        """
        return self.cam.get_exposure() / 1e6

    def set_auto_white_balance(self, toggle):
        """Enable/Disable auto white balance

        :param toggle: True if auto white balance
        :type toggle: bool
        """
        self.cam.set_param("auto_wb", toggle)

    def set_trigger_mode(self, external):
        """Set external trigger (edge rising).
        """
        if external:
            self.cam.set_gpi_mode("XI_GPI_TRIGGER")
            self.cam.set_trigger_source("XI_TRG_EDGE_RISING")
            self.cam.set_trigger_selector("XI_TRG_SEL_EXPOSURE_START")
        else:
            self.cam.set_trigger_source("XI_TRG_OFF")

    def set_vertical_skipping(self, factor):
        """Set vertical skipping
        """
        self.cam.set_downsampling_type("XI_SKIPPING")
        self.cam.set_decimation_vertical(factor)

    def set_limit_bandwidth(self, limit):
        """Enable limit bandwidth (useful if there are several cameras
        acquiring simultaneously).
        """
        if limit:
            self.cam.set_limit_bandwidth_mode("XI_ON")
        else:
            self.cam.set_limit_bandwidth_mode("XI_OFF")

    def get_trigger_mode(self):
        """Returns True if external, False if software
        """
        return self.cam.get_trigger_source() != "XI_TRG_OFF"

    def set_downsampling(self, binning):
        """Change image resolution by binning
        """
        self.cam.set_downsampling(binning)

    def set_roi(self, roiX0=0, roiY0=0, roiX1=0, roiY1=0):
        """This method sets the positions of the upper left corner (X0,Y0) and lower right
        (X1,Y1) corner of the ROI (region of interest) in pixels.
        """
        # Get bounds for width and height (with no offset)
        self.cam.set_offsetX(0)
        self.cam.set_offsetY(0)
        width_increment = self.cam.get_width_increment()
        width_minimum = self.cam.get_width_minimum()
        width_maximum = self.cam.get_width_maximum()
        height_increment = self.cam.get_height_increment()
        height_minimum = self.cam.get_height_minimum()
        height_maximum = self.cam.get_height_maximum()

        # Compute width and height
        width = roiX1 - roiX0
        height = roiY1 - roiY0

        # Check bounds for width and height
        if width < width_minimum:
            print("Minimum width is", width_minimum, "(got", width, ")")
            width = width_minimum
        elif width > width_maximum:
            print("Maximum width is", width_maximum, "(got", width, ")")
            width = width_maximum
        if height < height_minimum:
            print("Minimum height is", height_minimum, "(got", height, ")")
            height = height_minimum
        elif height > height_maximum:
            print("Maximum height is", height_maximum, "(got", height, ")")
            height = height_maximum
        if width % width_increment != 0:
            print("Rounding ROI width with", width_increment, "pixel increments")
            width -= width % width_increment
        if height % height_increment != 0:
            print("Rounding ROI height with", height_increment, "pixel increments")
            height -= height % height_increment

        # Set width and height
        self.cam.set_width(width)
        self.cam.set_height(height)

        # Get bounds for offsets
        offsetX_minimum = self.cam.get_offsetX_minimum()
        offsetX_maximum = self.cam.get_offsetX_maximum()
        offsetX_increment = self.cam.get_offsetX_increment()
        offsetY_minimum = self.cam.get_offsetY_minimum()
        offsetY_maximum = self.cam.get_offsetY_maximum()
        offsetY_increment = self.cam.get_offsetY_increment()

        # Compute offsets
        offset_x = roiX0
        offset_y = roiY0

        # Check bounds for offsets
        if offset_x < offsetX_minimum:
            print("Minimum x-offset is", offsetX_minimum)
            offset_x = offsetX_minimum
        elif offset_x > offsetX_maximum:
            print("Maximum x-offset is", offsetX_maximum)
            offset_x = offsetX_maximum
        if offset_x % offsetX_increment != 0:
            print("Rounding x-offset to", offsetX_increment, "pixel increments")
            offset_x -= offset_x % offsetX_increment
        if offset_y < offsetY_minimum:
            print("Minimum y-offset is", offsetY_minimum)
            offset_y = offsetY_minimum
        elif offset_y > offsetY_maximum:
            print("Maximum y-offset is", offsetY_maximum)
            offset_y = offsetY_maximum
        if offset_y % offsetY_increment != 0:
            print("Rounding y-offset to", offsetY_increment, "pixel increments")
            offset_y -= offset_y % offsetY_increment

        # Set offsets
        self.cam.set_offsetX(offset_x)
        self.cam.set_offsetY(offset_y)

        # Print final ROI
        print("ROI set to", offset_x, offset_y, offset_x + width, offset_y + height)

    def get_roi(self):
        """Get ROI
        """
        width = self.cam.get_width()
        height = self.cam.get_height()
        offset_x = self.cam.get_offsetX()
        offset_y = self.cam.get_offsetY()
        return offset_x, offset_x + width, offset_y, offset_y + height
        
    def set_frame_rate(self, framerate_fps):
        """Sets the framerate in frames per seconds
        """
        self.cam.set_acq_timing_mode("XI_ACQ_TIMING_MODE_FRAME_RATE_LIMIT")
        self.cam.set_framerate(framerate_fps)

    async def get_image(self, loop, image, timeout=5000):
        """Asynchronous version of xiapi.Camera.get_image method
        This function awaits for next image to be available in transport buffer.
        Color images are in BGR order (similar to OpenCV default).
        Attention: matplotlib expects RGB order.

        :param image: Image instance to copy image to
        :type image: :class:`xiapi.Image`
        :param timeout: timeout in milliseconds
        :type timeout: int
        """
        stat = await loop.run_in_executor(
            None,
            self.cam.device.xiGetImage,
            self.cam.handle,
            ctypes.wintypes.DWORD(timeout),
            ctypes.byref(image),
        )

        if stat == 10:
            raise CameraTimeout
        elif stat != 0:
            raise xiapi.Xi_error(stat)

    async def acquisition_async(
        self,
        num=np.inf,
        timeout=None,
        raw=False,
        initialising_cams=None,
        raise_on_timeout=True,
    ):
        """Concrete implementation of :meth:`pymanip.video.Camera.acquisition_async` for the Ximea camera.

        timeout in milliseconds.
        """
        loop = asyncio.get_event_loop()
        if timeout is None:
            timeout = max((5000, 5 * self.get_exposure_time() * 1000))
        img = xiapi.Image()
        self.cam.start_acquisition()
        try:
            count = 0
            while count < num:
                if (
                    count == 0
                    and initialising_cams is not None
                    and self in initialising_cams
                ):
                    initialising_cams.remove(self)
                try:
                    await self.get_image(loop, img, timeout)
                except CameraTimeout:
                    if raise_on_timeout:
                        raise
                    else:
                        stop_signal = yield None
                        if stop_signal:
                            break
                        else:
                            continue
                stop_signal = yield MetadataArray(
                    img.get_image_data_numpy(),  # no copy
                    metadata={
                        "counter": count,
                        "timestamp": img.tsSec + img.tsUSec * 1e-6,
                    },
                )
                count += 1
                if stop_signal:
                    break
        finally:
            self.cam.stop_acquisition()
        if stop_signal:
            yield True



"""

Async Stanford Resarch SR830 extension

"""

import fluidlab.instruments.amplifier.stanford_sr830 as fl_sr830
from pymanip.aioinstruments.aiofeatures import AsyncFloatValue
from pymanip.aioinstruments.aioiec60488 import AsyncIEC60488


class AsyncStanfordSR830SenseValue(AsyncFloatValue, fl_sr830.StanfordSR830SenseValue):
    async def aget(self):
        value = await super().aget()
        return self.sense_values[int(value)]

    async def set(self, value):
        await super().aset(self.sense_values.index(value))


class AsyncStanfordSR830TCValue(AsyncFloatValue, fl_sr830.StanfordSR830TCValue):
    async def aget(self):
        value = await super().aget()
        return self.tc_values[int(value)]

    async def aset(self, value):
        await super().aset(self.tc_values.index(value))


class AsyncStanfordSR830OffsetValue(AsyncFloatValue, fl_sr830.StanfordSR830OffsetValue):
    pass


class AsyncStanfordSR830ExpandValue(AsyncFloatValue, fl_sr830.StanfordSR830ExpandValue):
    pass


class AsyncStanfordSR830(AsyncIEC60488, fl_sr830.StanfordSR830):
    pass


afeatures = [
    AsyncFloatValue(
        "angle", doc="""Angle of the demodulated signal""", command_get="OUTP ? 4"
    ),
    AsyncFloatValue(
        "freq",
        doc="""Frequency of the excitation""",
        command_get="FREQ ?",
        command_set="FREQ",
    ),
    AsyncFloatValue(
        "mag", doc="""Magnitude of the demodulated signal""", command_get="OUTP ? 3"
    ),
    AsyncStanfordSR830OffsetValue(
        "offset",
        doc="""Channel offset in % of Sen (channel value is 1 or 2)""",
        command_get="OEXP ? {channel}",
        command_set="OEXP {channel},{value},0",
        channel_argument=True,
    ),
    AsyncStanfordSR830ExpandValue(
        "expand",
        doc="""Channel expand coefficient (channel value is 1 or 2)""",
        command_get="OEXP ? {channel}",
        command_set="OEXP {channel},{value},0",
        channel_argument=True,
    ),
    AsyncFloatValue(
        "reference_phase",
        doc="""Phase shift in degrees""",
        command_get="PHAS ?",
        command_set="PHAS",
    ),
    AsyncFloatValue(
        "vrms",
        doc="""RMS Voltage of the excitation""",
        command_get="SLVL ?",
        command_set="SLVL",
    ),
    AsyncFloatValue(
        "x", doc="""Demodulated value at theta=0""", command_get="OUTP ? 1"
    ),
    AsyncFloatValue(
        "y", doc="""Demodulated value at theta=90°""", command_get="OUTP ? 2"
    ),
    AsyncStanfordSR830SenseValue(),
    AsyncStanfordSR830TCValue(),
]

AsyncStanfordSR830._build_class_with_features(afeatures)



"""Arduino GBF Interface (:mod:`pymanip.aioinstruments.arduino_gbf`)
====================================================================

Interface for the Arduino GBF Interface (see the ino file in the Arduino
folder).

"""
import asyncio
import warnings
from time import monotonic, sleep

import numpy as np

from fluidlab.interfaces import PhysicalInterfaceType
from pymanip.aioinstruments.aiodrivers import AsyncDriver
from pymanip.aiodaq import (
    AcquisitionCard,
    TerminalConfig,
    TriggerConfig,
    TimeoutException,
)


class AsyncArduino(AsyncDriver, AcquisitionCard):
    default_physical_interface = PhysicalInterfaceType.Serial
    default_inter_params = {
        "baudrate": 9600,
        "bytesize": 8,
        "parity": "N",
        "stopbits": 1,
        "timeout": 5.0,
        "xonxoff": False,
        "rtscts": False,
        "dsrdtr": False,
        # "eol": "\r\n",
        # "autoremove_eol": True,
    }

    def __init__(self, interface, pin_output=13):
        AsyncDriver.__init__(self, interface)
        AcquisitionCard.__init__(self)
        self.pin_output = pin_output

    def __enter__(self):
        super().__enter__()
        while line := self.interface.readline().decode("ascii").strip():
            if line.startswith("GBF"):
                print("Connection established", line)
                break
            print(line)
        sleep(0.1)
        return self

    async def __aenter__(self):
        await super().__aenter__()
        start = monotonic()
        good = False
        while (monotonic() - start) < 5.0:
            identification = await self.interface.aread()
            identification = identification.decode("ascii").strip()
            if identification.startswith("GBF"):
                print("Connection to Arduino", identification)
                good = True
                break
            print(identification)
            await asyncio.sleep(0.1)
        if not good:
            raise RuntimeError("Timeout connecting to Arduino GBF")
        await asyncio.sleep(0.1)
        return self

    def close(self):
        pass

    def flush(self):
        while line := self.interface.readline().decode("ascii").strip():
            print(line)

    def add_channel(
        self, channel_name, terminal_config=TerminalConfig.RSE, voltage_range=5.0
    ):
        if hasattr(channel_name, "startswith") and channel_name.startswith("ai"):
            channel_name = channel_name[2:]
        pin_num = int(channel_name)
        self.channels = [str(pin_num)]
        if terminal_config != TerminalConfig.RSE:
            warnings.warn("Only RSE is supported")
        if voltage_range != 5.0:
            warnings.warn("Voltage range is fixed to 5.0 V")

    def configure_clock(self, sample_rate, samples_per_chan):
        self.sample_rate = sample_rate
        self.samples_per_chan = int(samples_per_chan)

    def configure_trigger(
        self,
        trigger_source=None,
        trigger_level=-5.0,
        trigger_config=TriggerConfig.EdgeRising,
    ):
        self.trigger_source = trigger_source
        self.trigger_level = trigger_level

    def start(self):
        if not hasattr(self, "trigger_level"):
            self.configure_trigger()
        delay_us = int(1e6 / self.sample_rate)
        self.interface.write(
            f"read_analog {self.channels[0]:} {self.samples_per_chan:d} {self.trigger_level:.1f} {delay_us:d}\n"
        )
        self.running = True

    async def stop(self):
        await self.interface.awrite("Abort\n")
        self.running = False

    async def read(self, tmo=None):
        self.reading = True
        start = monotonic()
        data = list()
        print("Start reading loop")
        while self.running:
            try:
                line = await self.interface.aread()
                line = line.decode("ascii").strip()
                if line.startswith("read_voltage_window FINISHED"):
                    print(line, "(breaking)")
                    break
                data.append(float(line))
            except ValueError:
                if not line:
                    print("no data (breaking)")
                    break
                print(line, "(continuing)")
                continue
            if tmo and monotonic() - start > tmo:
                raise TimeoutException()
        self.reading = False
        if data:
            return np.array(data)
        return None

    async def agenerate_pulse(self, npulses, delay_seconds, pin_output=None):
        pin = pin_output if pin_output is not None else self.pin_output
        delay_us = int(1e6 * delay_seconds)
        await self.interface.awrite(
            f"pulse {pin:d} {npulses:d} {delay_us:d}\n".encode("ascii")
        )
        while True:
            line = await self.interface.aread()
            line = line.decode("ascii").strip()
            if line:
                print(line)
                if line.startswith("single_pulse_generator FINISHED"):
                    break
            await asyncio.sleep(0.1)

    def configure_burst(self, freq, ncycles):
        """For compatibility with funcgen classes,

        .. code-block:: python

           arduino.configure_burst(freq, ncycles)
           arduino.trigger()

        is a equivalent to `arduino.generate_pulse(ncycles, 1/freq)`, except
        trigger() leave immediately.
        """
        self.freq = freq
        self.ncycles = ncycles

    def trigger(self, wait_completion=False):
        delay_us = int(1e6 / self.freq)
        self.interface.write(
            f"pulse {self.pin_output:d} {self.ncycles:d} {delay_us:d}\n"
        )
        while wait_completion and (
            line := self.interface.readline().decode("ascii").strip()
        ):
            print(line)
            if line.startswith("single_pulse_generator FINISHED"):
                break

    def configure_square(self, freq):
        delay_us = int(1e6 / freq)
        self.interface.write(f"continuous {self.pin_output:d} {delay_us:d}\n")

    def write_digital(self, pin_num, value):
        if pin_num < 2 or pin_num > 13:
            raise ValueError("Valid pins are between 2 and 13.")
        if value not in (0, 1):
            raise ValueError("Valid values are 0 or 1.")
        self.interface.write(f"write_digital {pin_num:d} {value:d}\n")
        start = monotonic()
        while line := self.interface.readline().decode("ascii").strip():
            print(line)
            if not line or line.startswith("write_digital FINISHED"):
                break
            if monotonic() - start > 5:
                print("Timeout reading from Arduino")
                break

    async def awrite_digital(self, pin_num, value):
        if pin_num < 2 or pin_num > 13:
            raise ValueError("Valid pins are between 2 and 13.")
        if value not in (0, 1):
            raise ValueError("Valid values are 0 or 1.")
        await self.interface.awrite(f"write_digital {pin_num:d} {value:d}\n")
        start = monotonic()
        while True:
            line = await self.interface.aread()
            line = line.decode("ascii").strip()
            print(line)
            if not line or line.startswith("digital_write"):
                break
            if monotonic() - start > 5:
                print("Timeout reading from Arduino")
                break



"""Andor reader module (:mod:`pymanip.video.andor.reader`)
==========================================================

This module implements simple pure-python reader for Andor DAT and SIF
files.

.. autoclass:: AndorAcquisitionReader
   :members:
   :private-members:

.. autoclass:: SIFFile
   :members:
   :private-members:

"""

import os
import io
import configparser
import glob
from datetime import datetime

import numpy as np


class AndorAcquisitionReader:
    """This class is a simple pure-python reader for Andor DAT spool files in a directory.

    :param acquisition_folder: the folder in which to read the DAT spool files
    :type acquisition_folder: str

    """

    def __init__(self, acquisition_folder):
        """Constructor method
        """
        config = configparser.ConfigParser()
        with io.open(
            os.path.join(acquisition_folder, "acquisitionmetadata.ini"),
            "r",
            encoding="utf-8-sig",
        ) as f:
            config.read_file(f)
            self.metadata = {
                "AOIWidth": int(config["data"]["aoiwidth"]),
                "AOIHeight": int(config["data"]["aoiheight"]),
                "AOIStride": int(config["data"]["aoistride"]),
                "PixelEncoding": str(config["data"]["pixelencoding"]),
                "ImageSizeBytes": int(config["data"]["imagesizebytes"]),
                "ImagesPerFile": int(config["multiimage"]["imagesperfile"]),
            }

            self.dt = np.uint16 if self.metadata["PixelEncoding"] == "Mono16" else None
            if not self.dt:
                raise ValueError("Unsupported pixel encoding")

            self.file_list = glob.glob(os.path.join(acquisition_folder, "*spool.dat"))
            self.file_list.sort()

    def images(self):
        """This generator method yields the images found in the folder.
        """
        for i in range(len(self)):
            yield self[i]

    def __getitem__(self, key):
        """This method returns the nth image in the folder.

        :param key: image number
        :type key: int

        """
        bname = os.path.basename(self.file_list[key])
        n = bname.find("spool")
        timestamp = int(bname[:n]) / 1e6
        with open(self.file_list[key], "rb") as f:
            _ = f.read(40)
            data = np.fromfile(f, dtype=self.dt).reshape(
                (self.metadata["AOIHeight"], self.metadata["AOIWidth"])
            )
        return timestamp, data[::-1, :]

    def __len__(self):
        """Number of images in folder.
        """
        return len(self.file_list)


class SIFFile:
    """This class implements a pure-python reader for Andor SIF files.

    :param filename: the SIF filename
    :type filename: str

    """

    def __init__(self, filename):
        """Constructor method
        """
        self.filename = filename

    def open(self):
        """This method opens the file.
        """
        self.f = open(self.filename, "rb")

    def close(self):
        """This method closes the file.
        """
        self.f.close()

    def __enter__(self):
        """Context manager enter method
        """
        self.open()
        self.read_header()
        return self

    def __exit__(self, type_, value, cb):
        """Context manager exit method
        """
        self.close()

    def read_header(self):
        """This method reads the header in the SIF file.
        """
        self.filetype = self.f.readline().decode("ascii").strip()
        if self.filetype != "Andor Technology Multi-Channel File":
            raise ValueError("Wrong file type")

        tokens = self.f.readline().decode("ascii").strip().split()
        self.version = int(tokens[0])
        self.signal_present = int(tokens[1])
        if self.version != 65538:
            raise ValueError("Unknown version")

        # InstaImage
        TInstaImage = dict()
        tokens = iter(self.f.readline().split())
        TInstaImage["version"] = int(next(tokens))
        TInstaImage["type"] = int(next(tokens))
        TInstaImage["active"] = int(next(tokens))
        TInstaImage["structure_version"] = int(next(tokens))
        TInstaImage["timedate"] = datetime.fromtimestamp(int(next(tokens)))
        TInstaImage["temperature"] = float(next(tokens))
        TInstaImage["head"] = next(tokens)
        TInstaImage["store_type"] = next(tokens)
        TInstaImage["data_type"] = next(tokens)
        TInstaImage["mode"] = next(tokens)
        TInstaImage["trigger_source"] = next(tokens)
        TInstaImage["trigger_level"] = float(next(tokens))
        TInstaImage["exposure_time"] = float(next(tokens))
        TInstaImage["delay"] = float(next(tokens))
        TInstaImage["integration_cycle_time"] = float(next(tokens))
        TInstaImage["no_integrations"] = int(next(tokens))
        TInstaImage["sync"] = next(tokens)
        TInstaImage["kinetic_cycle_time"] = float(next(tokens))
        TInstaImage["pixel_readout_time"] = float(next(tokens))
        TInstaImage["no_points"] = int(next(tokens))
        TInstaImage["fast_track_height"] = int(next(tokens))
        TInstaImage["gain"] = float(next(tokens))
        TInstaImage["gate_delay"] = float(next(tokens))
        TInstaImage["gate_width"] = float(next(tokens))
        TInstaImage["gate_step"] = float(next(tokens))
        TInstaImage["track_height"] = int(next(tokens))
        TInstaImage["series_length"] = int(next(tokens))
        TInstaImage["read_pattern"] = next(tokens)
        TInstaImage["shutter_delay"] = next(tokens)
        TInstaImage["st_centre_row"] = int(next(tokens))
        TInstaImage["mt_offset"] = int(next(tokens))
        TInstaImage["operation_mode"] = int(next(tokens))
        TInstaImage["FlipX"] = int(next(tokens))
        TInstaImage["FlipY"] = int(next(tokens))
        TInstaImage["Clock"] = int(next(tokens))
        TInstaImage["AClock"] = int(next(tokens))
        TInstaImage["MCP"] = int(next(tokens))
        TInstaImage["Prop"] = int(next(tokens))
        TInstaImage["IOC"] = int(next(tokens))
        TInstaImage["Freq"] = float(next(tokens))
        TInstaImage["VertClockAmp"] = float(next(tokens))
        TInstaImage["data_v_shift_speed"] = float(next(tokens))
        TInstaImage["OutputAmp"] = float(next(tokens))
        TInstaImage["PreAmpGain"] = float(next(tokens))
        TInstaImage["Serial"] = next(tokens)
        TInstaImage["NumPulses"] = int(next(tokens))
        TInstaImage["mFrameTransferAcqMode"] = int(next(tokens))
        if TInstaImage["version"] != 65550:
            TInstaImage["unstabilizedTemperature"] = float(next(tokens))
            TInstaImage["mBaselineClamp"] = float(next(tokens))
        TInstaImage["mPreScan"] = next(tokens)
        if TInstaImage["version"] in (65555, 65558, 65567):
            TInstaImage["mEMRealGain"] = next(tokens)
            TInstaImage["mBaselineOffset"] = next(tokens)
            TInstaImage["mSWVersion"] = next(tokens)
        if TInstaImage["version"] in (65558, 65567):
            TInstaImage["miGateMode"] = next(tokens)
            TInstaImage["mSWDllVer"] = next(tokens)
            TInstaImage["mSWDllRev"] = next(tokens)
            TInstaImage["mSWDllRel"] = next(tokens)
            TInstaImage["mSWDllBld"] = next(tokens)
        if TInstaImage["version"] == 65567:
            next(tokens)
            next(tokens)
        TInstaImage["len"] = int(next(tokens))

        TInstaImage["head_model"] = self.f.readline().decode("ascii").strip()
        tokens = iter(self.f.readline().decode("ascii").strip().split())

        TInstaImage["detector_format_x"] = next(tokens)
        TInstaImage["detector_format_y"] = next(tokens)
        TInstaImage["len2"] = next(tokens)

        TInstaImage["filename"] = self.f.readline().decode("ascii").strip()

        self.TInstaImage = TInstaImage

        # UserText
        TUserText = dict()
        tokens = iter(self.f.readline().decode("ascii").strip().split())
        TUserText["version"] = int(next(tokens))
        if TUserText["version"] != 65538:
            raise ValueError("Unknown TUserText version")
        TUserText["len"] = self.f.readline()
        TUserText["usertext"] = self.f.readline()
        self.f.readline()
        self.f.readline()
        self.f.readline()
        self.f.readline()
        self.f.readline()
        self.TUserText = TUserText

        # Shutter
        TShutter = dict()
        tokens = iter(self.f.readline().decode("ascii").strip().split())
        TShutter["version"] = int(next(tokens))
        if TShutter["version"] != 65538:
            raise ValueError("Unknown Shutter version")
        TShutter["type"] = next(tokens)
        TShutter["mode"] = next(tokens)
        TShutter["custom_bg_mode"] = next(tokens)
        TShutter["custom_mode"] = next(tokens)
        TShutter["custom_time"] = next(tokens)
        TShutter["opening_time"] = next(tokens)
        self.TShutter = TShutter

        # ShamrockSave
        TShamrockSave = dict()
        tokens = iter(self.f.readline().decode("ascii").strip().split())
        TShamrockSave["version"] = int(next(tokens))
        if TShamrockSave["version"] not in (65536, 65538, 65540):
            raise ValueError("Unknown ShamrockSave version")
        try:
            TShamrockSave["isActive"] = next(tokens)
            TShamrockSave["waveDrivePresent"] = next(tokens)
            TShamrockSave["wavelength"] = next(tokens)
            TShamrockSave["gratingTurretPresent"] = next(tokens)
            TShamrockSave["grating"] = next(tokens)
            TShamrockSave["gratingLines"] = next(tokens)
            TShamrockSave["gratingBlaze"] = next(tokens)
            TShamrockSave["slitPresent"] = next(tokens)
            TShamrockSave["slitWidth"] = next(tokens)
            TShamrockSave["flipperMirrorPresent"] = next(tokens)
            TShamrockSave["flipperPort"] = next(tokens)
            TShamrockSave["filterPresent"] = next(tokens)
            TShamrockSave["filterIndex"] = next(tokens)
            TShamrockSave["len"] = next(tokens)
            TShamrockSave["filterLabel"] = next(tokens)
            TShamrockSave["accessoryAttached"] = next(tokens)
            TShamrockSave["port1State"] = next(tokens)
            TShamrockSave["port2State"] = next(tokens)
            TShamrockSave["port3State"] = next(tokens)
            TShamrockSave["inputPortState"] = next(tokens)
            TShamrockSave["outputSlitPresent"] = next(tokens)
            TShamrockSave["outputSlitWidth"] = next(tokens)
        except StopIteration:
            pass

        for i in range(6):
            self.f.readline()

        self.TShamrockSave = TShamrockSave

        # SpectrographSave
        TSpectrographSave = dict()
        raw = self.f.readline().decode("ascii").strip()
        tokens = iter(raw.split())
        TSpectrographSave["version"] = int(next(tokens))
        TSpectrographSave["isActive"] = next(tokens)
        TSpectrographSave["wavelength"] = next(tokens)
        TSpectrographSave["gratingLines"] = next(tokens)
        tokens = iter(self.f.readline().decode("ascii").strip().split())
        TSpectrographSave["unspecified"] = next(tokens)
        TSpectrographSave["spectrographName"] = next(tokens)
        TSpectrographSave["unspecified2"] = next(tokens)
        self.f.readline()
        self.TSpectrographSave = TSpectrographSave

        # CalibImage
        TCalibImage = dict()
        raw = self.f.readline().decode("ascii").strip()
        tokens = iter(raw.split())
        TCalibImage["version"] = int(next(tokens))
        if TCalibImage["version"] != 65539:
            raise ValueError("Unknown TCalibImage version")

        # TImage
        while True:
            raw = self.f.readline()
            if raw.startswith(b"Pixel number65541"):
                raw = raw[12:]
                break

        TImage = dict()
        tokens = iter(raw.split())
        TImage["version"] = int(next(tokens))
        if TImage["version"] not in (65538, 65541):
            raise ValueError("Unknown TImage version")
        TImage["image_format.left"] = int(next(tokens))
        TImage["image_format.top"] = int(next(tokens))
        TImage["image_format.right"] = int(next(tokens))
        TImage["image_format.bottom"] = int(next(tokens))
        TImage["no_images"] = int(next(tokens))
        TImage["no_subimages"] = int(next(tokens))
        TImage["total_length"] = int(next(tokens))
        TImage["image_length"] = int(next(tokens))
        self.TImage = TImage

        # TSubImage
        TSubImage = dict()
        for subimage_index in range(TImage["no_subimages"]):
            raw = self.f.readline().decode("ascii").strip()
            tokens = iter(raw.split())
            TSubImage["version"] = int(next(tokens))
            if TSubImage["version"] != 65538:
                raise ValueError("Unknown SubImage version")
            TSubImage["left"] = int(next(tokens))
            TSubImage["top"] = int(next(tokens))
            TSubImage["right"] = int(next(tokens))
            TSubImage["bottom"] = int(next(tokens))
            TSubImage["vertical_bin"] = int(next(tokens))
            TSubImage["horizontal_bin"] = int(next(tokens))
            TSubImage["subimage_offset"] = int(next(tokens))
        self.TSubImage = TSubImage

        # timestamp
        self.timestamp = np.zeros((TImage["no_images"],))
        for i in range(TImage["no_images"]):
            self.timestamp[i] = int(self.f.readline().decode("ascii").strip())

        self.f.readline()

        for i in range(TImage["no_images"]):
            self.f.readline()

        self.datastart = self.f.tell()

        TImage["dimX"] = TImage["image_format.right"] - TImage["image_format.left"] + 1
        TImage["dimY"] = TImage["image_format.top"] - TImage["image_format.bottom"] + 1
        TSubImage["dimX"] = TSubImage["right"] - TSubImage["left"] + 1
        TSubImage["dimY"] = TSubImage["top"] - TSubImage["bottom"] + 1
        self.Npixels = TSubImage["dimX"] * TSubImage["dimY"]

    def read_nth_frame(self, n):
        """This method reads and returns the nth frame in the file.

        :param n: frame number to read
        :type n: int
        :return: frame
        :rtype: numpy.ndarray

        """
        self.f.seek(self.datastart + self.Npixels * n * 4)
        return self.read_frame()

    def read_frame(self):
        """The methods reads the next frame in file.

        :return: frame
        :rtype: numpy.ndarray
        """
        return np.fromfile(self.f, dtype=np.float32, count=self.Npixels).reshape(
            (self.TSubImage["dimY"], self.TSubImage["dimX"])
        )[::-1, :]

    def images(self):
        """This generator yields all the frames in the file.
        """
        self.f.seek(self.datastart)
        for ts in self.timestamp:
            yield ts, self.read_frame()

    def __len__(self):
        """Number of frames in file.
        """
        return self.timestamp.size



"""

Types used by NI System Configuration

"""

import ctypes
from pymanip.nisyscfg._lib.constants import NISysCfgResourceProperty as rp
from pymanip.nisyscfg._lib.constants import NISYSCFG_SIMPLE_STRING_LENGTH

NISysCfgEnumExpertHandle = ctypes.c_void_p
NISysCfgSessionHandle = ctypes.c_void_p
NISysCfgBool = ctypes.c_int
NISysCfgBusType = ctypes.c_int
NISysCfgHasDriverType = ctypes.c_int
NISysCfgIsPresentType = ctypes.c_int
NISysCfgTimestampUTC = ctypes.c_uint32 * 4
NISysCfgFirmwareUpdateMode = ctypes.c_int
NISysCfgAccessType = ctypes.c_int

# property: (attr_ctype, attr_ini, create_func, ref_func, enum_class)
NISysCfgResourcePropertyType = {
    rp.IsDevice: (NISysCfgBool, 0, None, ctypes.byref, None),
    rp.IsChassis: (NISysCfgBool, 0, None, ctypes.byref, None),
    rp.ConnectsToBusType: (NISysCfgBusType, 0, None, ctypes.byref, None),
    rp.VendorId: (ctypes.c_uint, 0, None, ctypes.byref, None),
    rp.VendorName: (
        ctypes.c_char_p,
        NISYSCFG_SIMPLE_STRING_LENGTH,
        ctypes.create_string_buffer,
        lambda x: x,
        None,
    ),
    rp.ProductId: (ctypes.c_uint, 0, None, ctypes.byref, None),
    rp.ProductName: (
        ctypes.c_char_p,
        NISYSCFG_SIMPLE_STRING_LENGTH,
        ctypes.create_string_buffer,
        lambda x: x,
        None,
    ),
    rp.SerialNumber: (
        ctypes.c_char_p,
        NISYSCFG_SIMPLE_STRING_LENGTH,
        ctypes.create_string_buffer,
        lambda x: x,
        None,
    ),
    rp.FirmwareRevision: (
        ctypes.c_char_p,
        NISYSCFG_SIMPLE_STRING_LENGTH,
        ctypes.create_string_buffer,
        lambda x: x,
        None,
    ),
    rp.IsNIProduct: (NISysCfgBool, 0, None, ctypes.byref, None),
    rp.IsSimulated: (NISysCfgBool, 0, None, ctypes.byref, None),
    rp.ConnectsToLinkName: (
        ctypes.c_char_p,
        NISYSCFG_SIMPLE_STRING_LENGTH,
        ctypes.create_string_buffer,
        lambda x: x,
        None,
    ),
    rp.HasDriver: (NISysCfgHasDriverType, 0, None, ctypes.byref, None),
    rp.IsPresent: (NISysCfgIsPresentType, 0, None, ctypes.byref, None),
    rp.SlotNumber: (ctypes.c_int, 0, None, ctypes.byref, None),
    rp.SupportsInternalCalibration: (NISysCfgBool, 0, None, ctypes.byref, None),
    rp.InternalCalibrationLastTime: (
        NISysCfgTimestampUTC,
        [0, 0, 0, 0],
        None,
        ctypes.byref,
        None,
    ),
    rp.InternalCalibrationLastTemp: (ctypes.c_double, 0.0, None, ctypes.byref, None),
    rp.SupportsExternalCalibration: (NISysCfgBool, 0, None, ctypes.byref, None),
    rp.ExternalCalibrationLastTemp: (ctypes.c_double, 0.0, None, ctypes.byref, None),
    rp.CalibrationComments: (
        ctypes.c_char_p,
        NISYSCFG_SIMPLE_STRING_LENGTH,
        ctypes.create_string_buffer,
        lambda x: x,
        None,
    ),
    rp.CurrentTemp: (ctypes.c_double, 0.0, None, ctypes.byref, None),
    rp.PxiPciBusNumber: (ctypes.c_uint, 0, None, ctypes.byref, None),
    rp.PxiPciDeviceNumber: (ctypes.c_uint, 0, None, ctypes.byref, None),
    rp.PxiPciFunctionNumber: (ctypes.c_uint, 0, None, ctypes.byref, None),
    rp.PxiPciLinkWidth: (ctypes.c_int, 0, None, ctypes.byref, None),
    rp.PxiPciMaxLinkWidth: (ctypes.c_int, 0, None, ctypes.byref, None),
    rp.UsbInterface: (ctypes.c_uint, 0, None, ctypes.byref, None),
    rp.TcpHostName: (
        ctypes.c_char_p,
        NISYSCFG_SIMPLE_STRING_LENGTH,
        ctypes.create_string_buffer,
        lambda x: x,
        None,
    ),
    rp.TcpMacAddress: (
        ctypes.c_char_p,
        NISYSCFG_SIMPLE_STRING_LENGTH,
        ctypes.create_string_buffer,
        lambda x: x,
        None,
    ),
    rp.TcpIpAddress: (
        ctypes.c_char_p,
        NISYSCFG_SIMPLE_STRING_LENGTH,
        ctypes.create_string_buffer,
        lambda x: x,
        None,
    ),
    rp.TcpDeviceClass: (
        ctypes.c_char_p,
        NISYSCFG_SIMPLE_STRING_LENGTH,
        ctypes.create_string_buffer,
        lambda x: x,
        None,
    ),
    rp.GpibPrimaryAddress: (ctypes.c_int, 0, None, ctypes.byref, None),
    rp.GpibSecondaryAddress: (ctypes.c_int, 0, None, ctypes.byref, None),
    rp.ProvidesBusType: (NISysCfgBusType, 0, None, ctypes.byref, None),
    rp.ProvidesLinkName: (
        ctypes.c_char_p,
        NISYSCFG_SIMPLE_STRING_LENGTH,
        ctypes.create_string_buffer,
        lambda x: x,
        None,
    ),
    rp.NumberOfSlots: (ctypes.c_int, 0, None, ctypes.byref, None),
    rp.SupportsFirmwareUpdate: (NISysCfgBool, 0, None, ctypes.byref, None),
    rp.FirmwareFilePattern: (
        ctypes.c_char_p,
        NISYSCFG_SIMPLE_STRING_LENGTH,
        ctypes.create_string_buffer,
        lambda x: x,
        None,
    ),
    rp.RecommendedCalibrationInterval: (ctypes.c_int, 0, None, ctypes.byref, None),
    rp.SupportsCalibrationWrite: (NISysCfgBool, 0, None, ctypes.byref, None),
    rp.HardwareRevision: (
        ctypes.c_char_p,
        NISYSCFG_SIMPLE_STRING_LENGTH,
        ctypes.create_string_buffer,
        lambda x: x,
        None,
    ),
    rp.CpuModelName: (
        ctypes.c_char_p,
        NISYSCFG_SIMPLE_STRING_LENGTH,
        ctypes.create_string_buffer,
        lambda x: x,
        None,
    ),
    rp.CpuSteppingRevision: (ctypes.c_int, 0, None, ctypes.byref, None),
    # rp.FirmwareUpdateMode: NISysCfgFirmwareUpdateMode,
    # rp.ExternalCalibrationLastTime: NISysCfgTimestampUTC,
    # rp.RecommendedNextCalibrationTime: NISysCfgTimestampUTC,
    rp.CalibrationCurrentPassword: (
        ctypes.c_char_p,
        NISYSCFG_SIMPLE_STRING_LENGTH,
        ctypes.create_string_buffer,
        lambda x: x,
        None,
    ),
    rp.CalibrationNewPassword: (
        ctypes.c_char_p,
        NISYSCFG_SIMPLE_STRING_LENGTH,
        ctypes.create_string_buffer,
        lambda x: x,
        None,
    ),
    # rp.SysCfgAccess: NISysCfgAccessType,
    # rp.AdapterType: NISysCfgAdapterType,
    rp.MacAddress: (
        ctypes.c_char_p,
        NISYSCFG_SIMPLE_STRING_LENGTH,
        ctypes.create_string_buffer,
        lambda x: x,
        None,
    ),
    # rp.AdapterMode: NISysCfgAdapterMode,
    # rp.TcpIpRequestMode: NISysCfgIpAddressMode,
    rp.TcpIpv4Address: (
        ctypes.c_char_p,
        NISYSCFG_SIMPLE_STRING_LENGTH,
        ctypes.create_string_buffer,
        lambda x: x,
        None,
    ),
    rp.TcpIpv4Subnet: (
        ctypes.c_char_p,
        NISYSCFG_SIMPLE_STRING_LENGTH,
        ctypes.create_string_buffer,
        lambda x: x,
        None,
    ),
    rp.TcpIpv4Gateway: (
        ctypes.c_char_p,
        NISYSCFG_SIMPLE_STRING_LENGTH,
        ctypes.create_string_buffer,
        lambda x: x,
        None,
    ),
    rp.TcpIpv4DnsServer: (
        ctypes.c_char_p,
        NISYSCFG_SIMPLE_STRING_LENGTH,
        ctypes.create_string_buffer,
        lambda x: x,
        None,
    ),
    # rp.TcpPreferredLinkSpeed: NISysCfgLinkSpeed,
    # rp.TcpCurrentLinkSpeed: NISysCfgLinkSpeed,
    # rp.TcpPacketDetection: NISysCfgPacketDetection,
    rp.TcpPollingInterval: (ctypes.c_uint, 0, None, ctypes.byref, None),
    rp.IsPrimaryAdapter: (NISysCfgBool, 0, None, ctypes.byref, None),
    rp.EtherCatMasterId: (ctypes.c_uint, 0, None, ctypes.byref, None),
    rp.EtherCatMasterRedundancy: (NISysCfgBool, 0, None, ctypes.byref, None),
    rp.WlanBssid: (
        ctypes.c_char_p,
        NISYSCFG_SIMPLE_STRING_LENGTH,
        ctypes.create_string_buffer,
        lambda x: x,
        None,
    ),
    rp.WlanCurrentLinkQuality: (ctypes.c_uint, 0, None, ctypes.byref, None),
    rp.WlanCurrentSsid: (
        ctypes.c_char_p,
        NISYSCFG_SIMPLE_STRING_LENGTH,
        ctypes.create_string_buffer,
        lambda x: x,
        None,
    ),
    # rp.WlanCurrentConnectionType: NISysCfgConnectionType,
    # rp.WlanCurrentSecurityType: NISysCfgSecurityType,
    # rp.WlanCurrentEapType: NISysCfgEapType,
    rp.WlanCountryCode: (ctypes.c_int, 0, None, ctypes.byref, None),
    rp.WlanChannelNumber: (ctypes.c_uint, 0, None, ctypes.byref, None),
    rp.WlanClientCertificate: (
        ctypes.c_char_p,
        NISYSCFG_SIMPLE_STRING_LENGTH,
        ctypes.create_string_buffer,
        lambda x: x,
        None,
    ),
    rp.WlanSecurityIdentity: (
        ctypes.c_char_p,
        NISYSCFG_SIMPLE_STRING_LENGTH,
        ctypes.create_string_buffer,
        lambda x: x,
        None,
    ),
    rp.WlanSecurityKey: (
        ctypes.c_char_p,
        NISYSCFG_SIMPLE_STRING_LENGTH,
        ctypes.create_string_buffer,
        lambda x: x,
        None,
    ),
    # rp.SystemStartTime: NISysCfgTimestampUTC ,
    # rp.CurrentTime: NISysCfgTimestampUTC,
    rp.TimeZone: (
        ctypes.c_char_p,
        NISYSCFG_SIMPLE_STRING_LENGTH,
        ctypes.create_string_buffer,
        lambda x: x,
        None,
    ),
    rp.UserDirectedSafeModeSwitch: (NISysCfgBool, 0, None, ctypes.byref, None),
    rp.ConsoleOutSwitch: (NISysCfgBool, 0, None, ctypes.byref, None),
    rp.IpResetSwitch: (NISysCfgBool, 0, None, ctypes.byref, None),
    rp.NumberOfDiscoveredAccessPoints: (ctypes.c_uint, 0, None, ctypes.byref, None),
    rp.NumberOfExperts: (ctypes.c_int, 0, None, ctypes.byref, None),
    rp.NumberOfServices: (ctypes.c_int, 0, None, ctypes.byref, None),
    rp.NumberOfAvailableFirmwareVersions: (ctypes.c_int, 0, None, ctypes.byref, None),
    rp.NumberOfCpus: (ctypes.c_int, 0, None, ctypes.byref, None),
    rp.NumberOfFans: (ctypes.c_int, 0, None, ctypes.byref, None),
    rp.NumberOfTemperatureSensors: (ctypes.c_int, 0, None, ctypes.byref, None),
    rp.NumberOfVoltageSensors: (ctypes.c_int, 0, None, ctypes.byref, None),
    rp.NumberOfUserLedIndicators: (ctypes.c_int, 0, None, ctypes.byref, None),
    rp.NumberOfUserSwitches: (ctypes.c_int, 0, None, ctypes.byref, None),
}



"""

Implements a live acquisition for NI cards

"""

import sys
import signal
import asyncio
import time
import math
from datetime import datetime

import numpy as np
import matplotlib.pyplot as plt
from matplotlib.widgets import Button, TextBox, CheckButtons, RadioButtons
import h5py

from pymanip.aiodaq import TerminalConfig as TC

Backends = dict()
try:
    from pymanip.aiodaq.daqmx import DAQmxSystem

    Backends["daqmx"] = DAQmxSystem
except ImportError:
    pass
try:
    from pymanip.aiodaq.scope import ScopeSystem

    Backends["scope"] = ScopeSystem
except ImportError:
    pass

from pymanip.aiodaq.arduino import ArduinoSystem

Backends["arduino"] = ArduinoSystem


class Oscillo:
    def __init__(
        self,
        channel_list,
        sampling=5e3,
        volt_range=10.0,
        trigger_level=None,
        trigsource=None,
        backend="daqmx",
        backend_args=None,
        N=1024,
    ):
        if backend not in Backends:
            raise ValueError(f"Backend {backend:} is not available.")
        self.backend = backend
        self.backend_args = backend_args
        self.channel_list = channel_list
        self.sampling = sampling
        self.volt_range = volt_range
        self.ignore_voltrange_submit = False
        self.trigger_level = trigger_level
        self.trigger_source = trigsource
        self.N = N
        plt.ion()
        self.fig = plt.figure()
        # left, bottom, width, height
        self.ax = self.fig.add_axes([0.1, 0.1, 0.7, 0.8])
        self.running = False
        self.last_trigged = 0
        self.ask_pause_acqui = False
        self.paused = False

        self.freq = None
        self.Pxx = None
        self.N_spectra = 0
        self.fig_spectrum = None
        self.hanning = True
        self.ac = True
        self.power_spectrum = True
        self.spectrum_unit = 1.0
        self.spectrum_unit_str = "V^2/Hz"
        self.system = None
        self.saved_spectra = list()

        self.fig_stats = None

        # Configure widgets
        left, bottom = 0.825, 0.825
        width, height = 0.15, 0.075
        vpad = 0.02

        ax_sampling = self.fig.add_axes([left, bottom, width, height])
        bottom -= height * 1.25 + 2 * vpad
        self.textbox_sampling = TextBox(
            ax_sampling, label="", initial=f"{sampling:.1e}"
        )
        self.textbox_sampling.on_submit(self.ask_sampling_change)
        _ = ax_sampling.text(0, 1.25, "Sampling")

        ax_enable_trigger = self.fig.add_axes([left, bottom, width, height])
        bottom -= height * 1.25 + 2 * vpad
        self.textbox_triggersource = TextBox(
            ax_enable_trigger, label="", initial="None"
        )
        self.textbox_triggersource.on_submit(self.ask_trigger_change)
        _ = ax_enable_trigger.text(0, 1.25, "Trigger")

        ax_triggerlevel = self.fig.add_axes([left, bottom, width, height])
        bottom -= height * 1.25 + 2 * vpad
        if trigger_level is not None:
            initial_value = f"{trigger_level:.2f}"
        else:
            initial_value = "1.0"
        self.textbox_triggerlevel = TextBox(
            ax_triggerlevel, label="", initial=initial_value
        )
        self.textbox_triggerlevel.on_submit(self.ask_trigger_change)
        _ = ax_triggerlevel.text(0, 1.25, "Level")

        ax_winsize = self.fig.add_axes([left, bottom, width, height])
        bottom -= height * 1.25 + 2 * vpad
        self.textbox_winsize = TextBox(ax_winsize, label="", initial=f"{self.N:d}")
        self.textbox_winsize.on_submit(self.ask_winsize_change)
        _ = ax_winsize.text(0, 1.25, "Win size")

        ax_voltrange = self.fig.add_axes([left, bottom, width, height])
        bottom -= height * 1.25 + 2 * vpad
        self.textbox_voltrange = TextBox(
            ax_voltrange, label="", initial=f"{self.volt_range:.1f}"
        )
        self.textbox_voltrange.on_submit(self.ask_voltrange_change)
        _ = ax_voltrange.text(0, 1.25, "Range")

        ax_start_stats = self.fig.add_axes([left, bottom, width, height])
        bottom -= height + vpad
        self.btn_start_stats = Button(ax_start_stats, label="Stats")
        self.btn_start_stats.on_clicked(self.start_stats)

        ax_start_spectrum = self.fig.add_axes([left, bottom, width, height])
        bottom -= height + vpad
        self.btn_start_spectrum = Button(ax_start_spectrum, label="FFT")
        self.btn_start_spectrum.on_clicked(self.start_spectrum)

    def clean_spectrum(self, *args):
        self.freq = None
        self.Pxx = None
        self.N_spectra = 0

    def start_stats(self, event):
        if self.fig_stats is None:
            self.fig_stats = dict()
            self.box_mean = dict()
            self.box_std = dict()
            self.box_min = dict()
            self.box_max = dict()
            self.box_freq = dict()
        nbox = 5
        height = 1.0 / (nbox + 1)
        padding = height / 4
        for chan in self.channel_list:
            if chan not in self.fig_stats:
                self.fig_stats[chan] = plt.figure(figsize=(2, 4))
                self.fig_stats[chan].canvas.set_window_title(chan)

                ax_mean = self.fig_stats[chan].add_axes(
                    [0.25, 9 * height / 2, 0.7, height - padding]
                )
                self.box_mean[chan] = TextBox(ax_mean, label="Mean", initial="")

                ax_std = self.fig_stats[chan].add_axes(
                    [0.25, 7 * height / 2, 0.7, height - padding]
                )
                self.box_std[chan] = TextBox(ax_std, label="Std", initial="")

                ax_min = self.fig_stats[chan].add_axes(
                    [0.25, 5 * height / 2, 0.7, height - padding]
                )
                self.box_min[chan] = TextBox(ax_min, label="Min", initial="")

                ax_max = self.fig_stats[chan].add_axes(
                    [0.25, 3 * height / 2, 0.7, height - padding]
                )
                self.box_max[chan] = TextBox(ax_max, label="Max", initial="")

                ax_freq = self.fig_stats[chan].add_axes(
                    [0.25, height / 2, 0.7, height - padding]
                )
                self.box_freq[chan] = TextBox(ax_freq, label="Freq", initial="")

    def start_spectrum(self, *args, **kwargs):
        if self.fig_spectrum is None:
            self.fig_spectrum = plt.figure()
            self.ax_spectrum = self.fig_spectrum.add_axes([0.1, 0.1, 0.7, 0.8])

            # Widgets
            ax_hanning = self.fig_spectrum.add_axes([0.825, 0.75, 0.15, 0.15])
            self.checkbox_hanning = CheckButtons(
                ax_hanning, ["Hanning", "AC"], [self.hanning, self.ac]
            )
            self.checkbox_hanning.on_clicked(self.ask_hanning_change)

            ax_spectrum_unit = self.fig_spectrum.add_axes([0.825, 0.51, 0.15, 0.25])
            self.radio_units = RadioButtons(
                ax_spectrum_unit,
                ["V^2/Hz", "V/sq(Hz)", "mV/sq(Hz)", "µV/sq(Hz)", "nV/sq(Hz)"],
            )
            self.radio_units.on_clicked(self.ask_spectrum_units_change)

            ax_restart = self.fig_spectrum.add_axes([0.825, 0.35, 0.15, 0.075])
            self.btn_restart = Button(ax_restart, label="Restart")
            self.btn_restart.on_clicked(self.clean_spectrum)

            ax_save_hold = self.fig_spectrum.add_axes([0.825, 0.25, 0.15, 0.075])
            self.btn_save_hold = Button(ax_save_hold, label="Hold&Save")
            self.btn_save_hold.on_clicked(self.save_hold)

        self.clean_spectrum()

    def save_hold(self, event):
        if self.N_spectra > 0:
            dt = datetime.now()
            filename = (
                f"pymanip_oscillo_{dt.year:}-{dt.month}-{dt.day}"
                f"_{dt.hour}-{dt.minute}-{dt.second}.hdf5"
            )
            bb = self.freq > 0
            with h5py.File(filename) as f:
                f.attrs["ts"] = dt.timestamp()
                f.attrs["N_spectra"] = self.N_spectra
                f.attrs["sampling"] = self.sampling
                f.attrs["volt_range"] = self.volt_range
                f.attrs["N"] = self.N
                f.create_dataset("freq", data=self.freq[bb])
                f.create_dataset("Pxx", data=self.Pxx[bb] / self.N_spectra)
            self.saved_spectra.append(
                {"freq": self.freq[bb], "Pxx": self.Pxx[bb] / self.N_spectra}
            )

    def ask_spectrum_units_change(self, event):
        power_spectrum_dict = {
            "V^2/Hz": True,
            "V/sq(Hz)": False,
            "mV/sq(Hz)": False,
            "µV/sq(Hz)": False,
            "nV/sq(Hz)": False,
        }
        spectrum_unit_dict = {
            "V^2/Hz": 1.0,
            "V/sq(Hz)": 1.0,
            "mV/sq(Hz)": 1e3,
            "µV/sq(Hz)": 1e6,
            "nV/sq(Hz)": 1e9,
        }
        if event in power_spectrum_dict:
            self.spectrum_unit_str = event
            self.power_spectrum = power_spectrum_dict[event]
            self.spectrum_unit = spectrum_unit_dict[event]

    async def winsize_change(self, new_N):
        await self.pause_acqui()
        old_N = self.N
        self.N = new_N
        self.clean_spectrum()
        self.figure_t_axis()
        try:
            self.system.configure_clock(self.sampling, self.N)
        except Exception as e:
            print(e)
            self.N = old_N
        await self.restart_acqui()

    def ask_winsize_change(self, label):
        try:
            new_N = int(label)
        except ValueError:
            print("winsize must be an integer")
            return
        loop = asyncio.get_event_loop()
        asyncio.ensure_future(self.winsize_change(new_N), loop=loop)

    def ask_trigger_change(self, label):
        changed = False
        trigger_source = self.textbox_triggersource.text
        possible_trigger_source = self.system.possible_trigger_channels()
        if trigger_source not in possible_trigger_source:
            print(f"{trigger_source:} is not an acceptable trigger source")
            print("Possible values:", possible_trigger_source)
            trigger_source = None
            self.textbox_triggersource.set_val("None")
        if trigger_source == "None":
            trigger_source = None
        if self.trigger_source != trigger_source:
            self.trigger_source = trigger_source
            changed = True
        if self.trigger_source is not None:
            try:
                trigger_level = float(self.textbox_triggerlevel.text)
            except ValueError:
                trigger_level = self.trigger_level
            val_str = f"{trigger_level:.2f}"
            self.textbox_triggerlevel.set_val(val_str)
            if trigger_level != self.trigger_level:
                self.trigger_level = trigger_level
                changed = True
        if changed:
            loop = asyncio.get_event_loop()
            asyncio.ensure_future(self.trigger_change(), loop=loop)

    async def pause_acqui(self):
        self.ask_pause_acqui = True
        await self.system.stop()
        while not self.paused:
            await asyncio.sleep(0.5)

    async def restart_acqui(self):
        self.ask_pause_acqui = False
        while self.paused:
            await asyncio.sleep(0.5)

    async def trigger_change(self):
        print("Awaiting pause")
        await self.pause_acqui()
        print("Acquisition paused")
        if self.trigger_level is not None:
            trigger_source = self.trigger_source
            trigger_level = self.trigger_level
            self.system.configure_trigger(trigger_source, trigger_level)
        else:
            self.system.configure_trigger(None)
        self.clean_spectrum()
        await self.restart_acqui()

    async def sampling_change(self):
        await self.pause_acqui()
        try:
            self.system.configure_clock(self.sampling, self.N)
        except Exception:
            print("Invalid sampling frequency")
            self.ask_sampling_change(self.system.samp_clk_max_rate)
            return
        if self.system.sample_rate != self.sampling:
            self.ask_sampling_change(self.system.sample_rate)
            return
        self.figure_t_axis()
        self.clean_spectrum()
        await self.restart_acqui()

    def ask_sampling_change(self, sampling):
        try:
            self.sampling = float(sampling)
            changed = True
        except ValueError:
            print("Wrong value:", sampling)
            changed = False
        self.textbox_sampling.set_val(f"{self.sampling:.1e}")
        if changed:
            loop = asyncio.get_event_loop()
            asyncio.ensure_future(self.sampling_change(), loop=loop)

    def ask_hanning_change(self, label):
        self.hanning, self.ac = self.checkbox_hanning.get_status()
        self.clean_spectrum()

    def figure_t_axis(self):
        self.t = np.arange(self.N) / self.sampling
        if self.t[-1] < 1:
            self.t *= 1000
            self.unit = "[ms]"
        else:
            self.unit = "[s]"

    async def run_gui(self):
        while self.running:
            if time.monotonic() - self.last_trigged > self.N / self.sampling:
                self.ax.set_title("Waiting for trigger")
            self.fig.canvas.start_event_loop(0.5)
            await asyncio.sleep(0.05)
            if not plt.fignum_exists(self.fig.number):
                self.running = False
            if self.fig_spectrum and not plt.fignum_exists(self.fig_spectrum.number):
                self.fig_spectrum = None
                self.freq = None
                self.Pxx = None
                self.N_spectra = 0
            if self.fig_stats is not None:
                for chan in list(self.fig_stats.keys()):
                    if not plt.fignum_exists(self.fig_stats[chan].number):
                        self.fig_stats.pop(chan)
                        self.box_mean.pop(chan)
                        self.box_std.pop(chan)
                        self.box_min.pop(chan)
                        self.box_max.pop(chan)
                        self.box_freq.pop(chan)
                    if not self.fig_stats:
                        self.fig_stats = None

    def ask_voltrange_change(self, new_range):
        if not self.ignore_voltrange_submit:
            try:
                new_range = float(new_range)
            except ValueError:
                print("Volt range must be a float")
                return
            loop = asyncio.get_event_loop()
            asyncio.ensure_future(self.voltrange_change(new_range), loop=loop)

    async def voltrange_change(self, new_range):
        await self.pause_acqui()
        self.volt_range = new_range
        self.clean_spectrum()
        self.create_system()
        await self.restart_acqui()
        actual_range = self.system.actual_ranges[0]
        print("actual_range =", actual_range)
        self.ignore_voltrange_submit = True
        self.textbox_voltrange.set_val(f"{actual_range:.1f}")
        self.ignore_voltrange_submit = False

    def create_system(self):
        if self.system is not None:
            self.system.close()
        if self.backend_args:
            self.system = Backends[self.backend](*self.backend_args)
        else:
            self.system = Backends[self.backend]()
        self.ai_channels = list()
        for chan in self.channel_list:
            ai_chan = self.system.add_channel(
                chan, terminal_config=TC.Diff, voltage_range=self.volt_range
            )
            self.ai_channels.append(ai_chan)
        self.system.configure_clock(self.sampling, self.N)
        self.textbox_sampling.set_val(f"{self.system.sample_rate:.1e}")
        if self.trigger_level is not None:
            trigger_source = self.channel_list[self.trigger_source]
            self.system.configure_trigger(
                trigger_source, trigger_level=self.trigger_level
            )
        else:
            self.system.configure_trigger(None)
        self.figure_t_axis()

    async def run_acqui(self):
        self.create_system()
        with self.system:
            while self.running:
                while self.ask_pause_acqui and self.running:
                    self.paused = True
                    await asyncio.sleep(0.5)
                self.paused = False
                if not self.running:
                    break
                self.system.start()
                data = await self.system.read()
                await self.system.stop()
                if data is None:
                    continue
                self.last_trigged = self.system.last_read
                self.ax.cla()
                if len(self.channel_list) == 1:
                    try:
                        self.ax.plot(self.t, data, "-")
                    except ValueError:
                        print("Inconsistent data")
                elif len(self.channel_list) > 1:
                    for d in data:
                        self.ax.plot(self.t, d, "-")
                if self.trigger_source not in (None, "Ext"):
                    self.ax.plot(
                        [self.t[0], self.t[-1]], [self.trigger_level] * 2, "g--"
                    )
                self.ax.set_xlim([self.t[0], self.t[-1]])
                self.ax.set_title("Trigged!")
                self.ax.set_xlabel("t " + self.unit)
                if self.fig_spectrum:
                    self.ax_spectrum.cla()
                    if self.saved_spectra:
                        for spectra in self.saved_spectra:
                            self.ax_spectrum.loglog(
                                spectra["freq"], spectra["Pxx"], "-"
                            )
                    if self.N_spectra == 0:
                        self.freq = np.fft.fftfreq(self.N, 1.0 / self.sampling)
                        bb = self.freq > 0
                        norm = math.pi * math.sqrt(self.N / self.sampling)
                        if self.hanning:
                            window = np.hanning(self.N)
                        else:
                            window = np.ones((self.N,))
                        if len(self.channel_list) == 1:
                            if self.ac:
                                m = np.mean(data)
                            else:
                                m = 0.0
                            self.Pxx = (
                                np.abs(np.fft.fft((data - m) * window) / norm) ** 2
                            )
                        else:
                            if self.ac:
                                ms = [np.mean(d) for d in data]
                            else:
                                ms = [0.0 for d in data]
                            self.Pxx = [
                                np.abs(np.fft.fft((d - m) * window) / norm) ** 2
                                for d, m in zip(data, ms)
                            ]
                        self.N_spectra = 1
                    else:
                        if len(self.channel_list) == 1:
                            if self.ac:
                                m = np.mean(data)
                            else:
                                m = 0.0
                            self.Pxx += (
                                np.abs(np.fft.fft((data - m) * window) / norm) ** 2
                            )
                        else:
                            if self.ac:
                                ms = [np.mean(d) for d in data]
                            else:
                                ms = [0.0 for d in data]
                            for p, d, m in zip(self.Pxx, data, ms):
                                p += np.abs(np.fft.fft((d - m) * window) / norm) ** 2
                        self.N_spectra += 1
                    if self.power_spectrum:

                        def process_spec(s):
                            return self.spectrum_unit * s

                    else:

                        def process_spec(s):
                            return self.spectrum_unit * np.sqrt(s)

                    if len(self.channel_list) == 1:
                        self.ax_spectrum.loglog(
                            self.freq[bb],
                            process_spec(self.Pxx[bb] / self.N_spectra),
                            "-",
                        )
                    else:
                        for p in self.Pxx:
                            self.ax_spectrum.loglog(
                                self.freq[bb], process_spec(p[bb] / self.N_spectra), "-"
                            )
                    self.ax_spectrum.set_xlabel("f [Hz]")
                    self.ax_spectrum.set_ylabel(self.spectrum_unit_str)
                    self.ax_spectrum.set_title(f"N = {self.N_spectra:d}")
                if self.fig_stats:
                    if len(self.channel_list) == 1:
                        list_data = [data]
                    else:
                        list_data = data
                    for chan, d in zip(self.channel_list, list_data):
                        if chan in self.fig_stats:
                            self.box_mean[chan].set_val("{:.5f}".format(np.mean(d)))
                            self.box_std[chan].set_val("{:.5f}".format(np.std(d)))
                            self.box_min[chan].set_val("{:.5f}".format(np.min(d)))
                            self.box_max[chan].set_val("{:.5f}".format(np.max(d)))
                            ff = np.fft.fftfreq(self.N, 1.0 / self.sampling)
                            pp = np.abs(np.fft.fft(d - np.mean(d)))
                            ii = np.argmax(pp)
                            self.box_freq[chan].set_val("{:.5f}".format(ff[ii]))

    def ask_exit(self, *args, **kwargs):
        self.running = False

    def run(self):
        loop = asyncio.get_event_loop()
        self.running = True
        if sys.platform == "win32":
            signal.signal(signal.SIGINT, self.ask_exit)
        else:
            for signame in ("SIGINT", "SIGTERM"):
                loop.add_signal_handler(getattr(signal, signame), self.ask_exit)

        loop.run_until_complete(asyncio.gather(self.run_gui(), self.run_acqui()))


if __name__ == "__main__":
    from pymanip.util.channel_selector import ChannelSelector

    chansel = ChannelSelector()
    chanlist = chansel.gui_select()
    oscillo = Oscillo(chanlist)
    oscillo.run()



"""Asynchronous IEC60488 instrument driver (:mod:`pymanip.aioinstruments.aioiec60488`)
======================================================================================

This module defines an asynchronous subclass of
:class:`fluidlab.instruments.iec60488.IEC60488`.

.. autoclass:: AsyncIEC60488
   :members:
   :private-members:

   .. autoattribute:: aclear_status

   .. autoattribute:: aquery_esr

   .. autoattribute:: aquery_stb

   .. autoattribute:: aquery_identification

   .. autoattribute:: areset_device

   .. autoattribute:: aperform_internal_test

   .. autoattribute:: await_till_completion_of_operations

   .. autoattribute:: aget_operation_complete_flag

   .. autoattribute:: await_continue


"""

import fluidlab.instruments.iec60488 as fl_iec60488
from pymanip.aioinstruments.aiofeatures import (
    AsyncWriteCommand,
    AsyncQueryCommand,
    AsyncRegisterValue,
)
from pymanip.aioinstruments.aiodrivers import AsyncDriver


class AsyncIEC60488(AsyncDriver, fl_iec60488.IEC60488):
    async def __aenter__(self):
        await super().__aenter__()
        identification = await self.aquery_identification()

        if isinstance(identification, str):
            identification = identification.strip()

        print(f"Initialization driver for device: {identification}")

        return self

    async def aquery_event_status_register(self):
        number = await self.aquery_esr()
        return self.status_enable_register.compute_dict_from_number(number)

    async def aquery_status_register(self):
        number = await self.aquery_stb()
        return self.status_enable_register.compute_dict_from_number(number)


afeatures = [
    # Reporting Commands
    AsyncWriteCommand("aclear_status", "Clears the data status structure", "*CLS"),
    AsyncRegisterValue(
        "event_status_enable_register",
        doc=(
            """Event status enable register

Used in the status and events reporting system.
"""
        ),
        command_set="*ESE",
        keys=fl_iec60488.EVENT_STATUS_BYTES,
    ),
    AsyncQueryCommand("aquery_esr", "Query the event status register", "*ESR?"),
    AsyncRegisterValue(
        "status_enable_register",
        doc=(
            """Status enable register

Used in the status reporting system.
"""
        ),
        command_set="*SRE",
        keys=fl_iec60488.EVENT_STATUS_BYTES,
    ),
    AsyncQueryCommand("aquery_stb", "Query the status register", "*STB?"),
    # Internal operation commands
    AsyncQueryCommand("aquery_identification", "Identification query", "*IDN?"),
    AsyncWriteCommand("areset_device", "Perform a device reset", "*RST"),
    AsyncQueryCommand("aperform_internal_test", "Perform internal self-test", "*TST?"),
    # Synchronization commands
    AsyncWriteCommand(
        "await_till_completion_of_operations",
        'Return "1" when all operation are completed',
        "*OPC",
    ),
    AsyncQueryCommand(
        "aget_operation_complete_flag", "Get operation complete flag", "*OPC?"
    ),
    AsyncQueryCommand("await_continue", "Wait to continue", "*WAI"),
]

AsyncIEC60488._build_class_with_features(afeatures)



"""

NI System Configuration DLL finding

"""

import ctypes
from ctypes.util import find_library
import os

libpath = find_library("nisyscfg")
if os.name == "posix":
    lib = ctypes.cdll.LoadLibrary(libpath)
elif os.name == "nt":
    lib = ctypes.windll.LoadLibrary(libpath)
